from __future__ import annotations

import json
from pathlib import Path

import pytest

pytest.importorskip("sqlalchemy")

from dc43_service_clients.data_quality import ValidationResult
from dc43_service_backends.governance.storage.sql import SQLGovernanceStore


@pytest.fixture()
def sql_engine(tmp_path: Path):
    from sqlalchemy import create_engine

    db_path = tmp_path / "governance.db"
    engine = create_engine(f"sqlite:///{db_path}")
    try:
        yield engine
    finally:
        engine.dispose()


def test_load_pipeline_activity_injects_dataset_version(sql_engine) -> None:
    store = SQLGovernanceStore(sql_engine)

    store._write_payload(  # type: ignore[attr-defined]
        store._activity,  # type: ignore[attr-defined]
        dataset_id="sales.orders",
        dataset_version="0.1.0",
        payload={
            "contract_id": "sales.orders",
            "contract_version": "0.1.0",
            "events": [
                {
                    "operation": "write",
                    "recorded_at": "2025-10-28T09:13:12.628161Z",
                }
            ],
        },
        extra={"updated_at": "2025-10-28T09:13:12.628161Z"},
    )

    activities = store.load_pipeline_activity(dataset_id="sales.orders")
    assert len(activities) == 1
    record = activities[0]
    assert record["dataset_id"] == "sales.orders"
    assert record["dataset_version"] == "0.1.0"
    assert record["contract_id"] == "sales.orders"
    assert record["contract_version"] == "0.1.0"
    assert record["events"] == [
        {
            "operation": "write",
            "recorded_at": "2025-10-28T09:13:12.628161Z",
        }
    ]

    single = store.load_pipeline_activity(
        dataset_id="sales.orders", dataset_version="0.1.0"
    )
    assert single == activities


def test_load_status_handles_duplicate_rows(sql_engine) -> None:
    payload_old = {
        "contract_id": "sales.orders",
        "contract_version": "0.1.0",
        "dataset_id": "sales.orders",
        "dataset_version": "0.1.0",
        "status": "warn",
        "reason": "initial run",
        "details": {"checks": ["row-count"]},
    }
    payload_new = dict(payload_old, status="ok", reason="latest run")

    with sql_engine.begin() as conn:
        conn.exec_driver_sql(
            """
            CREATE TABLE dq_status (
                dataset_id TEXT,
                dataset_version TEXT,
                contract_id TEXT,
                contract_version TEXT,
                payload TEXT,
                recorded_at TEXT
            )
            """
        )
        conn.exec_driver_sql(
            """
            INSERT INTO dq_status (
                dataset_id,
                dataset_version,
                contract_id,
                contract_version,
                payload,
                recorded_at
            ) VALUES (?, ?, ?, ?, ?, ?)
            """,
            (
                "sales.orders",
                "0.1.0",
                "sales.orders",
                "0.1.0",
                json.dumps(payload_old),
                "2024-01-01T00:00:00Z",
            ),
        )
        conn.exec_driver_sql(
            """
            INSERT INTO dq_status (
                dataset_id,
                dataset_version,
                contract_id,
                contract_version,
                payload,
                recorded_at
            ) VALUES (?, ?, ?, ?, ?, ?)
            """,
            (
                "sales.orders",
                "0.1.0",
                "sales.orders",
                "0.1.0",
                json.dumps(payload_new),
                "2024-01-02T00:00:00Z",
            ),
        )

    store = SQLGovernanceStore(sql_engine)

    result = store.load_status(
        contract_id="sales.orders",
        contract_version="0.1.0",
        dataset_id="sales.orders",
        dataset_version="0.1.0",
    )

    assert result is not None
    assert result.status == "ok"
    assert result.reason == "latest run"


def test_sql_store_extracts_metrics_from_details(sql_engine) -> None:
    store = SQLGovernanceStore(sql_engine)

    status = ValidationResult(status="ok")
    status.details = {"metrics": {"row_count": 3}}

    store.save_status(
        contract_id="sales.orders",
        contract_version="1.0.0",
        dataset_id="sales.orders",
        dataset_version="2024-06-01",
        status=status,
    )

    with sql_engine.begin() as conn:
        rows = conn.exec_driver_sql(
            "SELECT metric_key, metric_numeric_value FROM dq_metrics"
        ).fetchall()

    assert rows == [("row_count", 3.0)]


def test_sql_store_normalises_string_metrics(sql_engine) -> None:
    store = SQLGovernanceStore(sql_engine)

    status = ValidationResult(status="ok", metrics={"row_count": "12", "notes": "pass"})

    store.save_status(
        contract_id="sales.orders",
        contract_version="1.0.0",
        dataset_id="sales.orders",
        dataset_version="2024-06-01",
        status=status,
    )

    records = store.load_metrics(dataset_id="sales.orders")
    assert any(
        row["metric_key"] == "row_count" and row["metric_numeric_value"] == 12.0
        for row in records
    )
    assert any(
        row["metric_key"] == "notes" and row["metric_value"] == "pass"
        for row in records
    )


def test_sql_store_respects_explicit_metrics_table(sql_engine) -> None:
    store = SQLGovernanceStore(sql_engine, metrics_table="dq_status_metrics")

    assert store._metrics.name == "dq_status_metrics"  # type: ignore[attr-defined]


def test_sql_store_derives_metrics_table_from_status(sql_engine) -> None:
    store = SQLGovernanceStore(sql_engine, status_table="orders_dq_status")

    assert store._metrics.name == "orders_dq_metrics"  # type: ignore[attr-defined]


def test_sql_store_defaults_to_legacy_metrics_table(sql_engine) -> None:
    store = SQLGovernanceStore(sql_engine)

    assert store._metrics.name == "dq_metrics"  # type: ignore[attr-defined]


def test_sql_store_load_status_matrix_entries_filters(sql_engine) -> None:
    store = SQLGovernanceStore(sql_engine)

    store.save_status(
        contract_id="sales.orders",
        contract_version="1.0.0",
        dataset_id="sales.orders",
        dataset_version="2024-01-01",
        status=ValidationResult(status="ok"),
    )
    store.save_status(
        contract_id="sales.orders",
        contract_version="1.1.0",
        dataset_id="sales.orders",
        dataset_version="2024-01-02",
        status=ValidationResult(status="warn"),
    )

    entries = store.load_status_matrix_entries(
        dataset_id="sales.orders",
        dataset_versions=["2024-01-02"],
        contract_ids=["sales.orders"],
    )

    assert len(entries) == 1
    entry = entries[0]
    assert entry["dataset_version"] == "2024-01-02"
    status = entry["status"]
    assert isinstance(status, ValidationResult)
    assert status.status == "warn"



"""Interface for persisting governance artefacts."""

from __future__ import annotations

from typing import Mapping, Optional, Protocol, Sequence

from dc43_service_clients.data_quality import ValidationResult


class GovernanceStore(Protocol):
    """Persistence contract used by :class:`LocalGovernanceServiceBackend`."""

    def save_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        status: ValidationResult | None,
    ) -> None:
        """Persist the latest validation ``status`` for the dataset version."""

    def load_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
    ) -> ValidationResult | None:
        """Return the stored validation status for the dataset version."""

    def record_pipeline_event(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        event: Mapping[str, object],
        lineage_event: Mapping[str, object] | None = None,
    ) -> None:
        """Append ``event`` metadata to the pipeline activity log."""

    def list_datasets(self) -> Sequence[str]:
        """Return dataset identifiers recorded by the governance store."""

    def load_pipeline_activity(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        """Return pipeline activity entries for the dataset."""

    def link_dataset_contract(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: str,
        contract_version: str,
    ) -> None:
        """Persist an association between the dataset version and contract."""

    def get_linked_contract_version(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> str | None:
        """Return the contract reference linked to the dataset if any."""

    def load_metrics(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
        contract_id: Optional[str] = None,
        contract_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        """Return stored metric entries associated with the dataset."""

    def load_status_matrix_entries(
        self,
        *,
        dataset_id: str,
        dataset_versions: Sequence[str] | None = None,
        contract_ids: Sequence[str] | None = None,
    ) -> Sequence[Mapping[str, object]]:
        """Return persisted status rows for the dataset filtered by version and contract."""


__all__ = ["GovernanceStore"]



"""Compatibility shim exposing the relocated SQL governance store."""

from __future__ import annotations

from ..backend.stores import sql as _sql  # type: ignore
from ..backend.stores.sql import *  # type: ignore[F401,F403]

try:
    __all__ = list(_sql.__all__)
except AttributeError:
    __all__ = []


def __getattr__(name: str) -> object:
    """Delegate attribute access to the relocated module."""

    if name in _sql.__dict__:
        return _sql.__dict__[name]
    raise AttributeError(name)


def __dir__() -> list[str]:  # pragma: no cover - trivial passthrough
    """Combine the compatibility shim attributes with the target module ones."""

    return sorted(set(globals()) | set(dir(_sql)))



"""Bootstrap helpers for wiring governance extensions."""

from __future__ import annotations

from dataclasses import dataclass
from importlib import import_module
from typing import Callable, Iterable, Sequence, TYPE_CHECKING
import warnings

from dc43_service_backends.config import ServiceBackendsConfig

from .hooks import DatasetContractLinkHook

if TYPE_CHECKING:  # pragma: no cover - imported for type checking only
    from dc43_service_backends.contracts.backend.interface import ContractServiceBackend
else:  # pragma: no cover - keep runtime deps minimal
    ContractServiceBackend = object  # type: ignore[assignment]


@dataclass(slots=True)
class LinkHookContext:
    """Runtime services exposed to link-hook builders."""

    contract_service: "ContractServiceBackend | None" = None


LinkHookBuilder = Callable[
    [ServiceBackendsConfig, LinkHookContext | None],
    Sequence[DatasetContractLinkHook] | None,
]
LinkHookBuilderSpec = str

DEFAULT_LINK_HOOK_BUILDER_SPECS: tuple[LinkHookBuilderSpec, ...] = (
    "dc43_service_backends.governance.unity_catalog:build_link_hooks",
)


def load_link_hook_builder(spec: LinkHookBuilderSpec) -> LinkHookBuilder:
    """Import and return a link-hook builder referenced by ``spec``.

    The specification accepts the ``module:attribute`` format used by several
    Python plugin systems. When the delimiter is omitted, the right-most dot is
    treated as the attribute separator so dotted attribute paths remain
    supported.
    """

    if not spec:
        raise ValueError("link hook builder specification cannot be empty")

    module_name: str
    attribute_path: str
    if ":" in spec:
        module_name, attribute_path = spec.split(":", 1)
        module = import_module(module_name)
    else:
        parts = spec.split(".")
        if len(parts) < 2:
            raise ValueError(
                "link hook builder specification must include a module and attribute"
            )

        module = None
        for index in range(len(parts) - 1, 0, -1):
            module_name = ".".join(parts[:index])
            attribute_path = ".".join(parts[index:])
            try:
                module = import_module(module_name)
            except ModuleNotFoundError:
                continue
            else:
                break

        if module is None:
            raise ModuleNotFoundError(
                f"Unable to import link hook builder module for specification '{spec}'"
            )

    target: object = module
    for part in attribute_path.split("."):
        if not part:
            raise ValueError(
                f"invalid link hook builder attribute path in specification '{spec}'"
            )
        target = getattr(target, part)

    if not callable(target):  # pragma: no cover - defensive guard
        raise TypeError(f"link hook builder '{spec}' is not callable")

    return target  # type: ignore[return-value]


def _builder_specs_from_config(
    config: ServiceBackendsConfig,
    *,
    include_defaults: bool,
) -> tuple[LinkHookBuilderSpec, ...]:
    specs: list[LinkHookBuilderSpec] = []
    if include_defaults:
        specs.extend(DEFAULT_LINK_HOOK_BUILDER_SPECS)
    specs.extend(config.governance.dataset_contract_link_builders)

    ordered: list[LinkHookBuilderSpec] = []
    seen: set[LinkHookBuilderSpec] = set()
    for spec in specs:
        if not spec or spec in seen:
            continue
        seen.add(spec)
        ordered.append(spec)
    return tuple(ordered)


def build_dataset_contract_link_hooks(
    config: ServiceBackendsConfig,
    *,
    include_defaults: bool = True,
    extra_builders: Iterable[LinkHookBuilder] | None = None,
    context: LinkHookContext | None = None,
) -> tuple[DatasetContractLinkHook, ...]:
    """Assemble dataset-contract link hooks for the active configuration."""

    builders: list[LinkHookBuilder] = []
    for spec in _builder_specs_from_config(config, include_defaults=include_defaults):
        try:
            builder = load_link_hook_builder(spec)
        except Exception as exc:  # pragma: no cover - defensive guard
            warnings.warn(
                f"Dataset-contract link hook builder '{spec}' failed to load: {exc}",
                RuntimeWarning,
                stacklevel=2,
            )
            continue
        builders.append(builder)

    if extra_builders:
        builders.extend(extra_builders)

    hooks: list[DatasetContractLinkHook] = []
    for builder in builders:
        try:
            result = builder(config, context)
        except Exception as exc:  # pragma: no cover - defensive guard
            warnings.warn(
                f"Dataset-contract link hook builder failed: {exc}",
                RuntimeWarning,
                stacklevel=2,
            )
            continue

        if not result:
            continue

        hooks.extend(result)

    return tuple(hooks)


__all__ = [
    "build_dataset_contract_link_hooks",
    "DEFAULT_LINK_HOOK_BUILDER_SPECS",
    "LinkHookBuilder",
    "LinkHookBuilderSpec",
    "LinkHookContext",
    "load_link_hook_builder",
]



"""Delta Lake-backed governance persistence for Spark deployments."""

from __future__ import annotations

import json
import logging
import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import Mapping, Optional, Sequence, TYPE_CHECKING

from dc43_service_clients.data_quality import ValidationResult, coerce_details

from ._metrics import extract_metrics, normalise_metric_value
from ._table_names import (
    derive_related_table_basename,
    derive_related_table_name,
)
from .interface import GovernanceStore


logger = logging.getLogger(__name__)

if TYPE_CHECKING:  # pragma: no cover - import only for static typing
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.functions import Column

try:  # pragma: no cover - optional dependency guard
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.functions import col
    from pyspark.sql.types import (
        BooleanType,
        DoubleType,
        StringType,
        StructField,
        StructType,
    )
    from pyspark.sql.utils import AnalysisException
except ModuleNotFoundError as exc:  # pragma: no cover - allow import without pyspark
    raise ModuleNotFoundError(
        "pyspark is required for Delta governance storage. Install the optional "
        "dependencies with 'pip install dc43-service-backends[spark]'.",
    ) from exc

def _ensure_pyspark_is_remote_shim() -> None:
    """Expose ``is_remote`` for stubbed or legacy pyspark installs."""

    utils_module = sys.modules.get("pyspark.sql.utils")
    if utils_module is None or hasattr(utils_module, "is_remote"):
        return

    def _is_remote(*_: object, **__: object) -> bool:  # pragma: no cover - stub only
        return False

    setattr(utils_module, "is_remote", _is_remote)




_ensure_pyspark_is_remote_shim()


class DeltaGovernanceStore(GovernanceStore):
    """Persist governance artefacts to Delta tables."""

    def __init__(
        self,
        spark: SparkSession,
        *,
        base_path: str | Path | None = None,
        status_table: str | None = None,
        activity_table: str | None = None,
        link_table: str | None = None,
        metrics_table: str | None = None,
        bootstrap_tables: bool = True,
        log_sql: bool = False,
    ) -> None:
        if not base_path and not (status_table and activity_table and link_table):
            raise ValueError(
                "DeltaGovernanceStore requires either a base_path or explicit table names",
            )
        self._spark = spark
        self._base_path = Path(base_path).expanduser() if base_path else None
        self._status_table = status_table
        self._activity_table = activity_table
        self._link_table = link_table
        self._metrics_table = metrics_table
        self._log_sql = log_sql
        if not self._metrics_table and self._base_path is None:
            reference_table = status_table or activity_table or link_table
            if reference_table:
                self._metrics_table = self._derive_related_table_name(
                    reference_table, "metrics"
                )

        if bootstrap_tables:
            self.bootstrap()

    _derive_related_table_name = staticmethod(derive_related_table_name)
    _derive_related_table_basename = staticmethod(derive_related_table_basename)

    _STATUS_SCHEMA = StructType(
        [
            StructField("dataset_id", StringType(), False),
            StructField("dataset_version", StringType(), True),
            StructField("contract_id", StringType(), False),
            StructField("contract_version", StringType(), True),
            StructField("recorded_at", StringType(), False),
            StructField("deleted", BooleanType(), False),
            StructField("payload", StringType(), False),
        ]
    )
    _METRIC_SCHEMA = StructType(
        [
            StructField("dataset_id", StringType(), False),
            StructField("dataset_version", StringType(), True),
            StructField("contract_id", StringType(), False),
            StructField("contract_version", StringType(), True),
            StructField("status_recorded_at", StringType(), False),
            StructField("metric_key", StringType(), False),
            StructField("metric_value", StringType(), True),
            StructField("metric_numeric_value", DoubleType(), True),
        ]
    )
    _LINK_SCHEMA = StructType(
        [
            StructField("dataset_id", StringType(), False),
            StructField("dataset_version", StringType(), True),
            StructField("contract_id", StringType(), False),
            StructField("contract_version", StringType(), True),
            StructField("linked_at", StringType(), False),
        ]
    )
    _ACTIVITY_SCHEMA = StructType(
        [
            StructField("dataset_id", StringType(), False),
            StructField("dataset_version", StringType(), True),
            StructField("contract_id", StringType(), False),
            StructField("contract_version", StringType(), True),
            StructField("recorded_at", StringType(), False),
            StructField("payload", StringType(), False),
            StructField("pipeline_context", StringType(), True),
            StructField("lineage_event", StringType(), True),
        ]
    )

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------
    def _now(self) -> str:
        return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")

    def _table_path(self, name: str) -> str:
        assert self._base_path is not None
        path = self._base_path / name
        path.mkdir(parents=True, exist_ok=True)
        return str(path)

    @staticmethod
    def _escape_identifier(value: str) -> str:
        return value.replace("`", "``")

    @staticmethod
    def _escape_literal(value: str) -> str:
        return value.replace("'", "''")

    @staticmethod
    def _column_condition(column: str, value: Optional[str]) -> str:
        if value is None:
            return f"{column} IS NULL"
        escaped = DeltaGovernanceStore._escape_literal(str(value))
        return f"{column} = '{escaped}'"

    @staticmethod
    def _sql_target_identifier(table: str | None, folder: str | None) -> str | None:
        if table:
            return table
        if folder:
            return f"delta.`{DeltaGovernanceStore._escape_identifier(str(folder))}`"
        return None

    def _execute_sql(self, statement: str):  # pragma: no cover - logging shim
        if self._log_sql:
            logger.info("Spark SQL (governance): %s", statement.strip())
        return self._spark.sql(statement)

    def _purge_status_entries(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        table: str | None,
        folder: str | None,
    ) -> None:
        target = self._sql_target_identifier(table, folder)
        if not target:
            return
        condition = " AND ".join(
            (
                self._column_condition("dataset_id", dataset_id),
                self._column_condition("dataset_version", dataset_version),
            )
        )
        statement = f"DELETE FROM {target} WHERE {condition}"
        try:
            self._execute_sql(statement)
        except Exception:  # pragma: no cover - delta deletes may fail on legacy runtimes
            logger.exception(
                "Failed to purge existing status rows for %s@%s",
                dataset_id,
                dataset_version,
            )

    def _delta_folder_exists(self, folder: str | None) -> bool:
        if not folder:
            return False

        if hasattr(self._spark, "_jvm") and hasattr(self._spark, "_jsc"):
            try:
                jvm = self._spark._jvm
                jpath = jvm.org.apache.hadoop.fs.Path(str(folder))
                fs = jpath.getFileSystem(self._spark._jsc.hadoopConfiguration())
                delta_log_path = jvm.org.apache.hadoop.fs.Path(jpath, "_delta_log")
                return fs.exists(delta_log_path)
            except Exception:  # pragma: no cover - fall back to local checks
                pass

        path = Path(folder)
        return (path / "_delta_log").exists()

    def _table_exists(self, table: str | None) -> bool:
        if not table:
            return False

        if hasattr(self._spark, "sql"):
            parts = table.split(".")
            if len(parts) == 3:
                catalog, schema, name = parts

                def _escape(value: str) -> str:
                    return value.replace("'", "''")

                try:
                    query = (
                        "SELECT 1 FROM system.information_schema.tables "
                        f"WHERE table_catalog = '{_escape(catalog)}' "
                        f"AND table_schema = '{_escape(schema)}' "
                        f"AND table_name = '{_escape(name)}' "
                        "LIMIT 1"
                    )
                    rows = self._execute_sql(query).collect()
                    if rows:
                        return True
                except Exception:  # pragma: no cover - fall back to catalog lookup
                    pass

        return bool(self._spark.catalog.tableExists(table))


    def _ensure_delta_target(
        self,
        *,
        table: str | None,
        folder: str | None,
        schema: StructType,
    ) -> None:
        table_exists = self._table_exists(table)
        if table_exists:
            return
        folder_exists = self._delta_folder_exists(folder)
        if not table and folder_exists:
            return

        df = self._spark.createDataFrame([], schema)
        writer = (
            df.write.format("delta")
            .mode("overwrite")
            .option("overwriteSchema", "true")
        )
        if table:
            writer.saveAsTable(table)
        elif folder:
            Path(folder).mkdir(parents=True, exist_ok=True)
            writer.option("path", folder).save()
        else:  # pragma: no cover - defensive guard
            raise RuntimeError("DeltaGovernanceStore bootstrap target unspecified")

    def bootstrap(self) -> None:
        """Ensure Delta tables or folders exist for governance artefacts."""

        targets = (
            ("status", self._status_table, self._STATUS_SCHEMA),
            ("links", self._link_table, self._LINK_SCHEMA),
            ("activity", self._activity_table, self._ACTIVITY_SCHEMA),
            ("metrics", self._metrics_table, self._METRIC_SCHEMA),
        )
        for name, table, schema in targets:
            folder = self._table_path(name) if self._base_path else None
            if not table and not folder:
                continue
            self._ensure_delta_target(table=table, folder=folder, schema=schema)

    def _write(self, df: DataFrame, *, table: str | None, folder: str | None) -> None:
        writer = df.write.format("delta").mode("append")
        if table:
            writer.saveAsTable(table)
        elif folder:
            writer.option("path", folder).save()
        else:
            raise RuntimeError("DeltaGovernanceStore writer requires a table or folder")

    def _read(self, *, table: str | None, folder: str | None) -> DataFrame | None:
        try:
            if table:
                return self._spark.table(table)
            if folder:
                if not self._delta_folder_exists(folder):
                    return None
                return self._spark.read.format("delta").load(folder)
        except AnalysisException:
            return None
        return None

    # ------------------------------------------------------------------
    # Status persistence
    # ------------------------------------------------------------------
    def save_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        status: ValidationResult | None,
    ) -> None:
        recorded_at = self._now()
        folder = self._table_path("status") if self._base_path else None
        self._purge_status_entries(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            table=self._status_table,
            folder=folder,
        )
        details_payload = status.details if status else {}
        payload = {
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "contract_id": contract_id,
            "contract_version": contract_version,
            "recorded_at": recorded_at,
            "deleted": status is None,
            "payload": json.dumps(
                {
                    "status": status.status if status else "unknown",
                    "reason": status.reason if status else None,
                    "details": details_payload,
                    "contract_id": contract_id,
                    "contract_version": contract_version,
                    "dataset_id": dataset_id,
                    "dataset_version": dataset_version,
                    "deleted": status is None,
                }
            ),
        }
        df = self._spark.createDataFrame([payload], schema=self._STATUS_SCHEMA)
        self._write(df, table=self._status_table, folder=folder)

        if status is None:
            return

        metrics_map = extract_metrics(status)
        metrics_records = []
        for key, value in metrics_map.items():
            value_payload, numeric_value = normalise_metric_value(value)
            metrics_records.append(
                {
                    "dataset_id": dataset_id,
                    "dataset_version": dataset_version,
                    "contract_id": contract_id,
                    "contract_version": contract_version,
                    "status_recorded_at": recorded_at,
                    "metric_key": str(key),
                    "metric_value": value_payload,
                    "metric_numeric_value": numeric_value,
                }
            )

        if not metrics_records:
            return

        metrics_folder = self._table_path("metrics") if self._base_path else None
        if not self._metrics_table and not metrics_folder:
            return

        self._purge_status_entries(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            table=self._metrics_table,
            folder=metrics_folder,
        )

        metrics_df = self._spark.createDataFrame(metrics_records, self._METRIC_SCHEMA)
        self._write(metrics_df, table=self._metrics_table, folder=metrics_folder)

    def load_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
    ) -> ValidationResult | None:
        folder = self._table_path("status") if self._base_path else None
        df = self._read(table=self._status_table, folder=folder)
        if df is None:
            return None
        rows = (
            df.filter((col("dataset_id") == dataset_id) & (col("dataset_version") == dataset_version))
            .orderBy(col("recorded_at").desc())
            .limit(1)
            .collect()
        )
        if not rows:
            return None
        row = rows[0]
        record = json.loads(row.payload)
        if record.get("deleted"):
            return None
        linked = record.get("contract_id"), record.get("contract_version")
        if linked != (contract_id, contract_version):
            reason = (
                f"dataset linked to contract {linked[0]}:{linked[1]}"
                if all(linked)
                else "dataset linked to a different contract"
            )
            return ValidationResult(status="block", reason=reason, details=record)
        return ValidationResult(
            status=str(record.get("status", "unknown")),
            reason=str(record.get("reason")) if record.get("reason") else None,
            details=coerce_details(record.get("details")),
        )

    def load_status_matrix_entries(
        self,
        *,
        dataset_id: str,
        dataset_versions: Sequence[str] | None = None,
        contract_ids: Sequence[str] | None = None,
    ) -> Sequence[Mapping[str, object]]:
        folder = self._table_path("status") if self._base_path else None
        df = self._read(table=self._status_table, folder=folder)
        if df is None:
            return ()
        condition = col("dataset_id") == dataset_id
        versions = [str(value) for value in (dataset_versions or []) if str(value)]
        if versions:
            condition = condition & col("dataset_version").isin(versions)
        contracts = [str(value) for value in (contract_ids or []) if str(value)]
        if contracts:
            condition = condition & col("contract_id").isin(contracts)
        rows = df.filter(condition).collect()
        entries: list[Mapping[str, object]] = []
        for row in rows:
            if getattr(row, "deleted", False):
                continue
            record = json.loads(row.payload)
            status = ValidationResult(
                status=str(record.get("status", "unknown")),
                reason=str(record.get("reason")) if record.get("reason") else None,
                details=coerce_details(record.get("details")),
            )
            entries.append(
                {
                    "dataset_id": row.dataset_id,
                    "dataset_version": row.dataset_version,
                    "contract_id": row.contract_id,
                    "contract_version": row.contract_version,
                    "status": status,
                }
            )
        return tuple(entries)

    # ------------------------------------------------------------------
    # Dataset links
    # ------------------------------------------------------------------
    def link_dataset_contract(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: str,
        contract_version: str,
    ) -> None:
        payload = {
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "contract_id": contract_id,
            "contract_version": contract_version,
            "linked_at": self._now(),
        }
        df = self._spark.createDataFrame([payload], schema=self._LINK_SCHEMA)
        folder = self._table_path("links") if self._base_path else None
        
        self._purge_status_entries(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            table=self._link_table,
            folder=folder,
        )
        
        self._write(df, table=self._link_table, folder=folder)

    def get_linked_contract_version(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> str | None:
        folder = self._table_path("links") if self._base_path else None
        df = self._read(table=self._link_table, folder=folder)
        if df is None:
            return None
        condition = col("dataset_id") == dataset_id
        if dataset_version is not None:
            condition = condition & (col("dataset_version") == dataset_version)
        rows = df.filter(condition).orderBy(col("linked_at").desc()).limit(1).collect()
        if not rows:
            return None
        row = rows[0]
        cid = row.contract_id
        cver = row.contract_version
        if cid and cver:
            return f"{cid}:{cver}"
        return None

    # ------------------------------------------------------------------
    # Pipeline activity
    # ------------------------------------------------------------------
    def record_pipeline_event(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        event: Mapping[str, object],
        lineage_event: Mapping[str, object] | None = None,
    ) -> None:
        folder = self._table_path("activity") if self._base_path else None
        
        # Merge logic to avoid storing one row per streaming event and breaking load_pipeline_activity
        existing_records = self.load_pipeline_activity(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
        )
        record: dict[str, Any] = {"events": []}
        if existing_records:
            record = dict(existing_records[-1])
            
        events = list(record.get("events") or [])
        events.append(dict(event))
        if len(events) > 100:
            events = events[-100:]
        record["events"] = events
        
        if lineage_event is not None:
            record["lineage_event"] = dict(lineage_event)
        record["pipeline_context"] = dict(event.get("pipeline_context") or {})
        record["contract_id"] = contract_id
        record["contract_version"] = contract_version

        payload = {
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "contract_id": contract_id,
            "contract_version": contract_version,
            "recorded_at": self._now(),
            "payload": json.dumps(record),
            "pipeline_context": json.dumps(record["pipeline_context"]),
            "lineage_event": json.dumps(dict(lineage_event)) if lineage_event is not None else None,
        }
        df = self._spark.createDataFrame([payload], schema=self._ACTIVITY_SCHEMA)
        folder = self._table_path("activity") if self._base_path else None
        
        self._purge_status_entries(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            table=self._activity_table,
            folder=folder,
        )
        
        self._write(df, table=self._activity_table, folder=folder)

    def list_datasets(self) -> Sequence[str]:
        _ensure_pyspark_is_remote_shim()
        folder = self._table_path("activity") if self._base_path else None
        df = self._read(table=self._activity_table, folder=folder)
        if df is None:
            return []
        rows = df.select("dataset_id").distinct().orderBy(col("dataset_id")).collect()
        return [row.dataset_id for row in rows if getattr(row, "dataset_id", None)]

    def load_pipeline_activity(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        _ensure_pyspark_is_remote_shim()
        folder = self._table_path("activity") if self._base_path else None
        df = self._read(table=self._activity_table, folder=folder)
        if df is None:
            return []
        condition = col("dataset_id") == dataset_id
        if dataset_version is not None:
            condition = condition & (col("dataset_version") == dataset_version)
        rows = df.filter(condition).orderBy(col("recorded_at")).collect()
        aggregated: dict[str, dict[str, object]] = {}
        for row in rows:
            record = aggregated.setdefault(
                row.dataset_version,
                {
                    "dataset_id": row.dataset_id,
                    "dataset_version": row.dataset_version,
                    "contract_id": row.contract_id,
                    "contract_version": row.contract_version,
                    "events": [],
                },
            )
            try:
                event_payload = json.loads(row.payload)
            except json.JSONDecodeError:
                event_payload = {}
            if isinstance(event_payload, dict):
                record.setdefault("events", []).append(event_payload)
            lineage_raw = getattr(row, "lineage_event", None)
            if lineage_raw:
                try:
                    lineage_payload = json.loads(lineage_raw)
                except json.JSONDecodeError:
                    lineage_payload = None
                if isinstance(lineage_payload, dict):
                    record["lineage_event"] = lineage_payload
        ordered = sorted(aggregated.values(), key=lambda item: str(item.get("dataset_version", "")))
        if dataset_version is not None:
            return [ordered[0]] if ordered else []
        return ordered

    # ------------------------------------------------------------------
    # Metrics retrieval
    # ------------------------------------------------------------------
    def load_metrics(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
        contract_id: Optional[str] = None,
        contract_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        _ensure_pyspark_is_remote_shim()
        metrics_folder = self._table_path("metrics") if self._base_path else None
        df = self._read(table=self._metrics_table, folder=metrics_folder)
        if df is None:
            return []

        condition = col("dataset_id") == dataset_id
        if dataset_version is not None:
            condition = condition & (col("dataset_version") == dataset_version)
        if contract_id is not None:
            condition = condition & (col("contract_id") == contract_id)
        if contract_version is not None:
            condition = condition & (col("contract_version") == contract_version)

        rows = (
            df.filter(condition)
            .orderBy(col("status_recorded_at"), col("metric_key"))
            .collect()
        )
        entries: list[Mapping[str, object]] = []
        for row in rows:
            payload = row.asDict(recursive=False)
            numeric_value = payload.get("metric_numeric_value")
            if numeric_value is not None:
                try:
                    payload["metric_numeric_value"] = float(numeric_value)
                except (TypeError, ValueError):  # pragma: no cover - defensive
                    payload["metric_numeric_value"] = None
            entries.append(payload)
        return entries


__all__ = ["DeltaGovernanceStore"]



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List

from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

from dc43_service_backends.contracts.backend.local import LocalContractServiceBackend
from dc43_service_backends.contracts.backend.stores.interface import ContractStore
from dc43_service_backends.core.odcs import build_odcs


class InMemoryStore(ContractStore):
    def __init__(self) -> None:
        self._data: Dict[tuple[str, str], OpenDataContractStandard] = {}

    def put(self, contract: OpenDataContractStandard) -> None:  # type: ignore[override]
        self._data[(str(contract.id), str(contract.version))] = contract

    def get(self, contract_id: str, version: str) -> OpenDataContractStandard:  # type: ignore[override]
        return self._data[(contract_id, version)]

    def list_contracts(self) -> List[str]:  # type: ignore[override]
        return sorted({contract_id for contract_id, _ in self._data})

    def list_versions(self, contract_id: str) -> List[str]:  # type: ignore[override]
        return sorted(version for (cid, version) in self._data if cid == contract_id)


def make_contract(contract_id: str, version: str) -> OpenDataContractStandard:
    return build_odcs(
        contract_id=contract_id,
        version=version,
        kind="DatasetContract",
        api_version="3.0.2",
    )


def test_local_backend_delegates_to_store():
    store = InMemoryStore()
    backend = LocalContractServiceBackend(store)
    contract = make_contract("orders", "1.0.0")
    backend.put(contract)

    listing = backend.list_contracts()
    assert list(listing.items) == ["orders"]
    assert listing.total == 1
    assert backend.get("orders", "1.0.0") == contract
    assert backend.list_versions("orders") == ["1.0.0"]
    assert backend.latest("orders") == contract


def test_link_dataset_contract_keeps_contract_available():
    store = InMemoryStore()
    backend = LocalContractServiceBackend(store)
    backend.put(make_contract("orders", "1.0.0"))

    backend.link_dataset_contract(
        dataset_id="table.orders",
        dataset_version="current",
        contract_id="orders",
        contract_version="1.0.0",
    )

    assert backend.get("orders", "1.0.0").contract_id == "orders"
    assert backend.get_linked_contract_version(dataset_id="table.orders") is None



"""Filesystem-backed stub for governance-facing data-quality clients."""

from __future__ import annotations

import json
import logging
import os
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Mapping, Sequence

from dc43_service_backends.contracts.drafting import draft_from_validation_result
from dc43_service_backends.data_quality.backend.engine import evaluate_contract
from dc43_service_clients.data_quality import ValidationResult, coerce_details
from dc43_service_backends.core.odcs import contract_identity
from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

logger = logging.getLogger(__name__)


class StubDQClient:
    """Filesystem-backed stub for a DQ/DO service."""

    def __init__(self, base_path: str, *, block_on_violation: bool = True):
        self.base_path = base_path.rstrip("/")
        self.block_on_violation = block_on_violation
        logger.info("Initialized StubDQClient at %s", self.base_path)

    def _safe(self, s: str) -> str:
        """Return a filesystem-safe version of ``s``."""

        return "".join(ch if ch.isalnum() or ch in ("_", "-", ".") else "_" for ch in s)

    def _links_dir(self) -> str:
        d = os.path.join(self.base_path, "links")
        os.makedirs(d, exist_ok=True)
        return d

    def _links_path(self, dataset_id: str) -> str:
        return os.path.join(self._links_dir(), f"{self._safe(dataset_id)}.json")

    def _load_links(self, dataset_id: str) -> Dict[str, Any]:
        path = self._links_path(dataset_id)
        if not os.path.exists(path):
            return {"versions": {}}
        try:
            with open(path, "r", encoding="utf-8") as f:
                data = json.load(f)
        except (OSError, json.JSONDecodeError):
            return {"versions": {}}

        if not isinstance(data, dict):
            return {"versions": {}}

        # Upgrade legacy payloads that stored a single entry at the root.
        if "versions" not in data:
            entry = {
                key: data[key]
                for key in ("contract_id", "contract_version", "dataset_version", "linked_at")
                if key in data
            }
            upgraded: Dict[str, Any] = {"versions": {}}
            if entry.get("dataset_version"):
                upgraded["versions"][entry["dataset_version"]] = entry
                upgraded["latest"] = entry
            else:
                upgraded["latest"] = entry
            return upgraded

        versions = data.get("versions")
        if not isinstance(versions, dict):
            data["versions"] = {}
        return data

    def _status_path(self, dataset_id: str, dataset_version: str) -> str:
        d = os.path.join(self.base_path, "status", self._safe(dataset_id))
        os.makedirs(d, exist_ok=True)
        return os.path.join(d, f"{self._safe(str(dataset_version))}.json")

    def _activity_dir(self) -> str:
        d = os.path.join(self.base_path, "pipeline_activity")
        os.makedirs(d, exist_ok=True)
        return d

    def _activity_path(self, dataset_id: str) -> str:
        return os.path.join(self._activity_dir(), f"{self._safe(dataset_id)}.json")

    def _load_activity(self, dataset_id: str) -> Dict[str, Any]:
        path = self._activity_path(dataset_id)
        if not os.path.exists(path):
            return {"dataset_id": dataset_id, "versions": {}}
        try:
            with open(path, "r", encoding="utf-8") as f:
                payload = json.load(f)
        except (OSError, json.JSONDecodeError):
            return {"dataset_id": dataset_id, "versions": {}}

        if not isinstance(payload, dict):
            return {"dataset_id": dataset_id, "versions": {}}
        versions = payload.get("versions")
        if not isinstance(versions, dict):
            payload["versions"] = {}
        return payload

    def get_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
    ) -> ValidationResult:
        path = self._status_path(dataset_id, dataset_version)
        logger.debug("Fetching DQ status from %s", path)
        if not os.path.exists(path):
            return ValidationResult(status="unknown", reason="no-status-for-version")
        with open(path, "r", encoding="utf-8") as f:
            data = json.load(f)
        link = self.get_linked_contract_version(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
        )
        if link and link != f"{contract_id}:{contract_version}":
            return ValidationResult(status="block", reason=f"dataset linked to contract {link}", details=data)
        return ValidationResult(status=data.get("status", "warn"), reason=data.get("reason"), details=data.get("details", {}))

    def submit_metrics(
        self,
        *,
        contract: OpenDataContractStandard,
        dataset_id: str,
        dataset_version: str,
        metrics: Dict[str, Any],
    ) -> ValidationResult:
        schema_payload = metrics.get("schema") if isinstance(metrics, dict) else None
        metric_values = metrics.copy() if isinstance(metrics, dict) else {}
        if "schema" in metric_values:
            metric_values = dict(metric_values)
            metric_values.pop("schema", None)

        evaluation = evaluate_contract(
            contract,
            schema=schema_payload if isinstance(schema_payload, Mapping) else None,
            metrics=metric_values,
            strict_types=True,
            allow_extra_columns=True,
            expectation_severity="error",
        )

        blocking = self.block_on_violation
        violations = 0
        for k, v in metric_values.items():
            if k.startswith("violations.") or k.startswith("query."):
                if isinstance(v, (int, float)):
                    violations += int(v)

        if evaluation.errors:
            status = "block" if blocking else "warn"
        elif evaluation.warnings and blocking:
            status = "warn"
        else:
            status = "ok"

        details = coerce_details(evaluation.details)
        if not details:
            details = {
                "errors": list(evaluation.errors),
                "warnings": list(evaluation.warnings),
                "metrics": dict(evaluation.metrics),
                "schema": dict(evaluation.schema),
            }
        details["violations"] = violations

        path = self._status_path(dataset_id, dataset_version)
        logger.info("Persisting DQ status %s for %s@%s to %s", status, dataset_id, dataset_version, path)
        contract_id_value, contract_version_value = contract_identity(contract)
        recorded_at = datetime.now(timezone.utc).isoformat()
        payload = {
            "status": status,
            "details": details,
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "contract_id": contract_id_value,
            "contract_version": contract_version_value,
            "recorded_at": recorded_at,
        }
        with open(path, "w", encoding="utf-8") as f:
            json.dump(payload, f)

        self.link_dataset_contract(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            contract_id=contract_id_value,
            contract_version=contract_version_value,
        )
        return ValidationResult(status=status, details=details)

    def load_status_matrix_entries(
        self,
        *,
        dataset_id: str,
        dataset_versions: Sequence[str] | None = None,
        contract_ids: Sequence[str] | None = None,
    ) -> Sequence[Mapping[str, object]]:
        versions = {str(value) for value in (dataset_versions or []) if str(value)}
        contracts = {str(value) for value in (contract_ids or []) if str(value)}
        status_dir = os.path.join(self.base_path, "status", self._safe(dataset_id))
        if not os.path.isdir(status_dir):
            return ()
        entries: list[Mapping[str, object]] = []
        for name in os.listdir(status_dir):
            path = os.path.join(status_dir, name)
            try:
                with open(path, "r", encoding="utf-8") as handle:
                    payload = json.load(handle)
            except (OSError, json.JSONDecodeError):
                continue
            dataset_version = str(payload.get("dataset_version") or "")
            if versions and dataset_version not in versions:
                continue
            contract_id = str(payload.get("contract_id") or "")
            if contracts and contract_id not in contracts:
                continue
            contract_version = str(payload.get("contract_version") or "")
            status = ValidationResult(
                status=str(payload.get("status", "unknown")),
                reason=str(payload.get("reason")) if payload.get("reason") else None,
                details=coerce_details(payload.get("details")),
            )
            entries.append(
                {
                    "dataset_id": dataset_id,
                    "dataset_version": dataset_version,
                    "contract_id": contract_id,
                    "contract_version": contract_version,
                    "status": status,
                }
            )
        return tuple(entries)

    def record_pipeline_activity(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: Optional[str],
        contract_version: Optional[str],
        activity: Mapping[str, Any],
    ) -> None:
        payload = self._load_activity(dataset_id)
        versions = payload.setdefault("versions", {})
        entry = versions.get(dataset_version)
        if not isinstance(entry, dict):
            entry = {"dataset_version": dataset_version, "events": []}

        events: List[Dict[str, Any]]
        raw_events = entry.get("events")
        if isinstance(raw_events, list):
            events = [event for event in raw_events if isinstance(event, dict)]
        else:
            events = []

        event = dict(activity)
        event.setdefault("recorded_at", datetime.now(timezone.utc).isoformat())
        context_payload = event.get("pipeline_context")
        if isinstance(context_payload, Mapping):
            event["pipeline_context"] = dict(context_payload)
        elif context_payload is None:
            event["pipeline_context"] = {}

        events.append(event)

        entry["dataset_version"] = dataset_version
        entry["contract_id"] = contract_id
        entry["contract_version"] = contract_version
        entry["events"] = events
        versions[dataset_version] = entry
        payload["dataset_id"] = dataset_id
        payload["latest"] = entry

        path = self._activity_path(dataset_id)
        with open(path, "w", encoding="utf-8") as f:
            json.dump(payload, f)

    def get_pipeline_activity(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
        include_status: bool = False,
    ) -> List[Dict[str, Any]]:
        payload = self._load_activity(dataset_id)
        versions = payload.get("versions")
        if not isinstance(versions, Mapping):
            return []

        def _normalise_entry(key: str, value: Mapping[str, Any]) -> Dict[str, Any]:
            record = dict(value)
            record.setdefault("dataset_version", key)
            events = record.get("events")
            if not isinstance(events, list):
                record["events"] = []
            else:
                record["events"] = [event for event in events if isinstance(event, dict)]
            return record

        if dataset_version is not None:
            entry = versions.get(dataset_version)
            if isinstance(entry, Mapping):
                return [_normalise_entry(dataset_version, entry)]
            # Attempt to match unsanitised versions stored as values
            for candidate_version, candidate in versions.items():
                if (
                    isinstance(candidate, Mapping)
                    and candidate.get("dataset_version") == dataset_version
                ):
                    return [_normalise_entry(candidate_version, candidate)]
            return []

        records: List[Dict[str, Any]] = []
        for key, value in versions.items():
            if isinstance(value, Mapping):
                records.append(_normalise_entry(str(key), value))

        def _sort_key(entry: Mapping[str, Any]) -> tuple[int, str]:
            version_value = str(entry.get("dataset_version", ""))
            recorded_at = ""
            events = entry.get("events")
            if isinstance(events, list) and events:
                last = events[-1]
                if isinstance(last, Mapping):
                    recorded_at = str(last.get("recorded_at", ""))
            return (0 if recorded_at else 1, recorded_at or version_value)

        records.sort(key=_sort_key)
        return records

    def propose_draft(
        self,
        *,
        validation: ValidationResult,
        base_contract: OpenDataContractStandard,
        bump: str = "minor",
        dataset_id: Optional[str] = None,
        dataset_version: Optional[str] = None,
        data_format: Optional[str] = None,
        dq_feedback: Optional[Mapping[str, Any]] = None,
        draft_context: Optional[Mapping[str, Any]] = None,
    ) -> Optional[OpenDataContractStandard]:
        """Produce a draft contract using the engine helper."""

        return draft_from_validation_result(
            validation=validation,
            base_contract=base_contract,
            bump=bump,
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            data_format=data_format,
            dq_feedback=dq_feedback,
            draft_context=draft_context,
        )

    def link_dataset_contract(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: str,
        contract_version: str,
    ) -> None:
        path = self._links_path(dataset_id)
        logger.info(
            "Linking dataset %s@%s to contract %s:%s at %s",
            dataset_id,
            dataset_version,
            contract_id,
            contract_version,
            path,
        )
        payload = self._load_links(dataset_id)
        linked_at = datetime.now(timezone.utc).isoformat()
        entry = {
            "contract_id": contract_id,
            "contract_version": contract_version,
            "dataset_version": dataset_version,
            "linked_at": linked_at,
        }
        versions = payload.setdefault("versions", {})
        versions[dataset_version] = entry
        payload["latest"] = entry
        with open(path, "w", encoding="utf-8") as f:
            json.dump(payload, f)

    def get_linked_contract_version(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> Optional[str]:
        payload = self._load_links(dataset_id)

        def _entry_to_link(entry: Mapping[str, Any] | None) -> Optional[str]:
            if not isinstance(entry, Mapping):
                return None
            contract_id_value = entry.get("contract_id")
            contract_version_value = entry.get("contract_version")
            if contract_id_value and contract_version_value:
                link_value = f"{contract_id_value}:{contract_version_value}"
                logger.debug("Found contract link for %s -> %s", dataset_id, link_value)
                return link_value
            return None

        versions = payload.get("versions") if isinstance(payload, Mapping) else None
        if dataset_version and isinstance(versions, Mapping):
            entry = versions.get(dataset_version)
            if entry is None:
                # Attempt to match on stored dataset version values when keys were sanitised.
                for candidate in versions.values():
                    if (
                        isinstance(candidate, Mapping)
                        and candidate.get("dataset_version") == dataset_version
                    ):
                        entry = candidate
                        break
            link = _entry_to_link(entry)
            if link:
                return link

        latest = payload.get("latest") if isinstance(payload, Mapping) else None
        link = _entry_to_link(latest)
        if link:
            return link

        # Compatibility with legacy payloads that stored the entry at the root.
        return _entry_to_link(payload if isinstance(payload, Mapping) else None)


__all__ = ["StubDQClient"]



from __future__ import annotations

import json
import re
import sys
import types
from pathlib import Path
from types import SimpleNamespace


def _maybe_install_pyspark_stub() -> None:
    """Provide a lightweight pyspark replacement when the real package is absent."""

    try:
        import pyspark  # noqa: F401  # pragma: no cover - import check only
        from pyspark import SparkContext  # type: ignore  # pragma: no cover - import check only
        from pyspark.sql import SparkSession  # type: ignore  # pragma: no cover - import check only
        from pyspark.sql.functions import col  # type: ignore  # pragma: no cover - import check only
        from pyspark.sql.types import StructType  # type: ignore  # pragma: no cover - import check only
        from pyspark.sql.utils import AnalysisException  # type: ignore  # pragma: no cover - import check only
    except ModuleNotFoundError:
        pass
    else:
        # A functional pyspark installation is available; no stub required.
        return

    for name in [
        "pyspark",
        "pyspark.sql",
        "pyspark.sql.functions",
        "pyspark.sql.types",
        "pyspark.sql.utils",
    ]:
        sys.modules.pop(name, None)

    pyspark = types.ModuleType("pyspark")
    sql_module = types.ModuleType("pyspark.sql")

    class _StubSparkContext:  # pragma: no cover - attribute container only
        """Fallback stand-in so imports expecting ``SparkContext`` succeed."""

        _active_spark_context: "_StubSparkContext | None" = None

        def __init__(self) -> None:  # pragma: no cover - behaviour unused in tests
            _StubSparkContext._active_spark_context = self

        @classmethod
        def getOrCreate(cls) -> "_StubSparkContext":  # pragma: no cover - helper parity
            if cls._active_spark_context is None:
                cls._active_spark_context = cls()
            return cls._active_spark_context

    class _StubSparkSession:  # pragma: no cover - attribute container only
        _active_session: "_StubSparkSession | None" = None

        @classmethod
        def getActiveSession(cls) -> "_StubSparkSession | None":
            return cls._active_session

        @classmethod
        def setActiveSession(
            cls, session: "_StubSparkSession | None"
        ) -> None:
            cls._active_session = session

        class builder:  # pragma: no cover - builder parity for defensive code paths
            @staticmethod
            def master(_: str) -> "_StubSparkSession.builder":
                return _StubSparkSession.builder()

            @staticmethod
            def appName(_: str) -> "_StubSparkSession.builder":
                return _StubSparkSession.builder()

            @staticmethod
            def getOrCreate() -> "_StubSparkSession":
                session = _StubSparkSession()
                _StubSparkSession.setActiveSession(session)
                _StubSparkContext.getOrCreate()
                return session

    class _StubDataFrame:  # pragma: no cover - attribute container only
        pass

    sql_module.SparkSession = _StubSparkSession
    sql_module.DataFrame = _StubDataFrame

    functions_module = types.ModuleType("pyspark.sql.functions")

    class _StubColumn:  # pragma: no cover - attribute container only
        def __init__(self, name: str) -> None:
            self.name = name

        def __eq__(self, _: object) -> "_StubColumn":
            return self

        def __and__(self, _: object) -> "_StubColumn":
            return self

        def desc(self) -> "_StubColumn":  # pragma: no cover - behaviour ignored in stub
            return self

    def _col(name: str) -> _StubColumn:
        return _StubColumn(name)

    functions_module.col = _col

    types_module = types.ModuleType("pyspark.sql.types")

    class _StubStructField(tuple):
        def __new__(cls, name: str, data_type: object, nullable: bool) -> "_StubStructField":
            return tuple.__new__(cls, (name, data_type, nullable))

    class _StubStructType(list):
        def __init__(self, fields: list[_StubStructField]):
            super().__init__(fields)

    def _boolean_type() -> str:
        return "boolean"

    def _double_type() -> str:
        return "double"

    def _string_type() -> str:
        return "string"

    types_module.StructField = _StubStructField
    types_module.StructType = _StubStructType
    types_module.BooleanType = _boolean_type
    types_module.DoubleType = _double_type
    types_module.StringType = _string_type

    utils_module = types.ModuleType("pyspark.sql.utils")

    class _StubAnalysisException(Exception):
        pass

    utils_module.AnalysisException = _StubAnalysisException

    pyspark.SparkContext = _StubSparkContext  # type: ignore[attr-defined]
    pyspark.sql = sql_module
    sql_module.functions = functions_module
    sql_module.types = types_module
    sql_module.utils = utils_module

    sys.modules["pyspark"] = pyspark
    sys.modules["pyspark.sql"] = sql_module
    sys.modules["pyspark.sql.functions"] = functions_module
    sys.modules["pyspark.sql.types"] = types_module
    sys.modules["pyspark.sql.utils"] = utils_module


_maybe_install_pyspark_stub()

class _StubFileSystem:
    def __init__(self, existing_paths: set[str]) -> None:
        self._existing_paths = existing_paths

    def exists(self, path: object) -> bool:
        if hasattr(path, "toString"):
            return path.toString() in self._existing_paths
        return False


class _StubHadoopPath:
    def __init__(self, factory: "_StubPathFactory", value: str) -> None:
        self._factory = factory
        self._value = value

    def getFileSystem(self, _conf: object) -> _StubFileSystem:
        return self._factory.filesystem

    def toString(self) -> str:  # pragma: no cover - convenience for debugging
        return self._value


class _StubPathFactory:
    def __init__(self, existing_paths: set[str]) -> None:
        self.filesystem = _StubFileSystem(existing_paths)

    def __call__(self, base: object, child: str | None = None) -> _StubHadoopPath:
        if child is None:
            value = self._to_string(base)
        else:
            prefix = self._to_string(base).rstrip("/")
            value = f"{prefix}/{child}"
        return _StubHadoopPath(self, value)

    @staticmethod
    def _to_string(value: object) -> str:
        if hasattr(value, "toString"):
            return value.toString()
        return str(value)

from dc43_service_backends.governance.storage.delta import DeltaGovernanceStore
from dc43_service_clients.data_quality import ValidationResult


class _RecordingWriter:
    def __init__(self, records: list[dict[str, object]]) -> None:
        self._records = records
        self._format: str | None = None
        self._mode: str | None = None
        self._options: dict[str, object] = {}

    def format(self, name: str) -> "_RecordingWriter":
        self._format = name
        return self

    def mode(self, name: str) -> "_RecordingWriter":
        self._mode = name
        return self

    def option(self, key: str, value: object) -> "_RecordingWriter":
        self._options[key] = value
        return self

    def saveAsTable(self, table: str) -> None:
        self._records.append(
            {
                "type": "table",
                "target": table,
                "format": self._format,
                "mode": self._mode,
                "options": dict(self._options),
            }
        )

    def save(self) -> None:
        self._records.append(
            {
                "type": "path",
                "target": self._options.get("path"),
                "format": self._format,
                "mode": self._mode,
                "options": dict(self._options),
            }
        )


class _RecordingDataFrame:
    def __init__(self, records: list[dict[str, object]]) -> None:
        self._records = records

    @property
    def write(self) -> _RecordingWriter:
        return _RecordingWriter(self._records)


class _StubSqlResult:
    def __init__(self, exists: bool) -> None:
        self._exists = exists

    def collect(self) -> list[object]:
        if self._exists:
            return [SimpleNamespace(result=1)]
        return []


class _StubSpark:
    def __init__(
        self,
        *,
        existing_tables: set[str] | None = None,
        filesystem_paths: set[str] | None = None,
        information_schema: set[tuple[str, str, str]] | None = None,
    ) -> None:
        self._records: list[dict[str, object]] = []
        self._dataframes: list[dict[str, object]] = []
        self._schemas: list[object] = []
        self._existing_tables = existing_tables or set()
        self._information_schema = information_schema or set()
        self.catalog = SimpleNamespace(tableExists=self._table_exists)
        self._sql_queries: list[str] = []
        if filesystem_paths is not None:
            path_factory = _StubPathFactory(filesystem_paths)
            self._jvm = SimpleNamespace(
                org=SimpleNamespace(
                    apache=SimpleNamespace(
                        hadoop=SimpleNamespace(fs=SimpleNamespace(Path=path_factory))
                    )
                )
            )
            self._jsc = SimpleNamespace(hadoopConfiguration=lambda: SimpleNamespace())

    def _table_exists(self, name: str) -> bool:
        return name in self._existing_tables

    @property
    def records(self) -> list[dict[str, object]]:
        return self._records

    @property
    def schemas(self) -> list[object]:
        return self._schemas

    @property
    def sql_queries(self) -> list[str]:
        return self._sql_queries

    @property
    def dataframes(self) -> list[dict[str, object]]:
        return self._dataframes

    def createDataFrame(self, data: list[object], schema: object | None = None) -> _RecordingDataFrame:
        self._schemas.append(schema)
        self._dataframes.append({"data": list(data), "schema": schema})
        return _RecordingDataFrame(self._records)

    def sql(self, query: str) -> _StubSqlResult:
        self._sql_queries.append(query)
        if "system.information_schema.tables" not in query:
            return _StubSqlResult(False)
        matches = dict(re.findall(r"table_(catalog|schema|name)\s*=\s*'([^']*)'", query))
        catalog = matches.get("catalog")
        schema = matches.get("schema")
        name = matches.get("name")
        exists = (
            catalog is not None
            and schema is not None
            and name is not None
            and (catalog, schema, name) in self._information_schema
        )
        return _StubSqlResult(exists)


def _tables(records: list[dict[str, object]]) -> list[dict[str, object]]:
    return [entry for entry in records if entry["type"] == "table"]


def _paths(records: list[dict[str, object]]) -> list[dict[str, object]]:
    return [entry for entry in records if entry["type"] == "path"]


def test_bootstrap_creates_missing_delta_tables() -> None:
    spark = _StubSpark()
    DeltaGovernanceStore(
        spark,
        status_table="analytics.governance.status",
        activity_table="analytics.governance.activity",
        link_table="analytics.governance.links",
        metrics_table="analytics.governance.metrics",
    )

    tables = _tables(spark.records)
    assert {entry["target"] for entry in tables} == {
        "analytics.governance.status",
        "analytics.governance.activity",
        "analytics.governance.links",
        "analytics.governance.metrics",
    }
    for entry in tables:
        assert entry["format"] == "delta"
        assert entry["mode"] == "overwrite"
        assert entry["options"].get("overwriteSchema") == "true"


def test_bootstrap_derives_metrics_table_name() -> None:
    spark = _StubSpark()
    DeltaGovernanceStore(
        spark,
        status_table="analytics.governance.status",
        activity_table="analytics.governance.activity",
        link_table="analytics.governance.links",
    )

    tables = _tables(spark.records)
    assert {entry["target"] for entry in tables} == {
        "analytics.governance.status",
        "analytics.governance.activity",
        "analytics.governance.links",
        "analytics.governance.status_metrics",
    }
    for entry in tables:
        assert entry["format"] == "delta"
        assert entry["mode"] == "overwrite"
        assert entry["options"].get("overwriteSchema") == "true"


def test_bootstrap_derives_metrics_table_name_from_status_suffix() -> None:
    spark = _StubSpark()
    DeltaGovernanceStore(
        spark,
        status_table="analytics.governance.dq_status",
        activity_table="analytics.governance.activity",
        link_table="analytics.governance.links",
    )

    tables = _tables(spark.records)
    assert {entry["target"] for entry in tables} == {
        "analytics.governance.dq_status",
        "analytics.governance.activity",
        "analytics.governance.links",
        "analytics.governance.dq_metrics",
    }
    for entry in tables:
        assert entry["format"] == "delta"
        assert entry["mode"] == "overwrite"
        assert entry["options"].get("overwriteSchema") == "true"


def test_bootstrap_skips_existing_tables() -> None:
    spark = _StubSpark(existing_tables={"analytics.governance.status"})
    DeltaGovernanceStore(
        spark,
        status_table="analytics.governance.status",
        activity_table="analytics.governance.activity",
        link_table="analytics.governance.links",
        metrics_table="analytics.governance.metrics",
    )

    tables = _tables(spark.records)
    assert {entry["target"] for entry in tables} == {
        "analytics.governance.activity",
        "analytics.governance.links",
        "analytics.governance.metrics",
    }


def test_bootstrap_skips_tables_detected_via_information_schema() -> None:
    spark = _StubSpark(
        information_schema={("analytics", "governance", "status")}
    )
    DeltaGovernanceStore(
        spark,
        status_table="analytics.governance.status",
        activity_table="analytics.governance.activity",
        link_table="analytics.governance.links",
        metrics_table="analytics.governance.metrics",
    )

    tables = _tables(spark.records)
    assert {entry["target"] for entry in tables} == {
        "analytics.governance.activity",
        "analytics.governance.links",
        "analytics.governance.metrics",
    }
    assert any(
        "information_schema.tables" in query for query in spark.sql_queries
    )


def test_bootstrap_initialises_delta_folders(tmp_path: Path) -> None:
    spark = _StubSpark()
    DeltaGovernanceStore(spark, base_path=tmp_path)

    paths = _paths(spark.records)
    assert {Path(str(entry["target"])).name for entry in paths} == {
        "status",
        "links",
        "activity",
        "metrics",
    }
    for entry in paths:
        assert entry["format"] == "delta"
        assert entry["mode"] == "overwrite"
        assert entry["options"].get("overwriteSchema") == "true"


def test_bootstrap_skips_existing_delta_folders(tmp_path: Path) -> None:
    status_log = tmp_path / "status" / "_delta_log"
    status_log.mkdir(parents=True)

    spark = _StubSpark()
    DeltaGovernanceStore(spark, base_path=tmp_path)

    paths = _paths(spark.records)
    assert {Path(str(entry["target"])).name for entry in paths} == {
        "links",
        "activity",
        "metrics",
    }


def test_remote_delta_folder_detection_uses_spark_filesystem(tmp_path: Path) -> None:
    spark = _StubSpark(filesystem_paths={"s3://bucket/status/_delta_log"})
    store = DeltaGovernanceStore(spark, base_path=tmp_path, bootstrap_tables=False)

    store._ensure_delta_target(
        table=None,
        folder="s3://bucket/status",
        schema=DeltaGovernanceStore._STATUS_SCHEMA,
    )

    assert not spark.records


def test_save_status_records_metrics_entries(tmp_path: Path) -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(spark, base_path=tmp_path)
    store._now = lambda: "2024-03-01T00:00:00Z"  # type: ignore[assignment]

    status = ValidationResult(status="ok", metrics={"violations.total": 3, "summary": {"passed": 10}})

    store.save_status(
        contract_id="contracts",
        contract_version="1.0.0",
        dataset_id="orders",
        dataset_version="2024-02-29",
        status=status,
    )

    metrics_frames = [
        frame
        for frame in spark.dataframes
        if frame["schema"] is DeltaGovernanceStore._METRIC_SCHEMA and frame["data"]
    ]
    assert len(metrics_frames) == 1
    rows = metrics_frames[0]["data"]
    assert {row["metric_key"] for row in rows} == {"violations.total", "summary"}
    total_row = next(row for row in rows if row["metric_key"] == "violations.total")
    assert total_row["metric_numeric_value"] == 3.0
    assert total_row["status_recorded_at"] == "2024-03-01T00:00:00Z"
    summary_row = next(row for row in rows if row["metric_key"] == "summary")
    assert json.loads(summary_row["metric_value"] or "{}") == {"passed": 10}
    assert summary_row["metric_numeric_value"] is None


def test_save_status_records_metrics_from_details_payload(tmp_path: Path) -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(spark, base_path=tmp_path)
    store._now = lambda: "2024-05-01T00:00:00Z"  # type: ignore[assignment]

    status = ValidationResult(status="ok")
    status.details = {"metrics": {"row_count": 3}}

    store.save_status(
        contract_id="contracts",
        contract_version="1.0.0",
        dataset_id="orders",
        dataset_version="2024-05-01",
        status=status,
    )

    metrics_frames = [
        frame
        for frame in spark.dataframes
        if frame["schema"] is DeltaGovernanceStore._METRIC_SCHEMA and frame["data"]
    ]
    assert len(metrics_frames) == 1
    rows = metrics_frames[0]["data"]
    assert rows[0]["metric_key"] == "row_count"
    assert rows[0]["metric_numeric_value"] == 3.0


def test_save_status_appends_metrics_to_table_target() -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(
        spark,
        status_table="analytics.status",
        activity_table="analytics.activity",
        link_table="analytics.links",
        metrics_table="analytics.metrics",
    )
    store._now = lambda: "2024-04-05T12:00:00Z"  # type: ignore[assignment]

    status = ValidationResult(status="ok", metrics={"violations.total": 1})

    store.save_status(
        contract_id="contracts",
        contract_version="2.0.0",
        dataset_id="orders",
        dataset_version="2024-04-01",
        status=status,
    )

    metric_writes = [entry for entry in spark.records if entry["target"] == "analytics.metrics"]
    assert metric_writes
    assert all(entry["type"] == "table" for entry in metric_writes)


def test_save_status_deletes_existing_rows_for_table_target() -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(
        spark,
        status_table="analytics.status",
        activity_table="analytics.activity",
        link_table="analytics.links",
        metrics_table="analytics.metrics",
    )

    status = ValidationResult(status="ok")

    store.save_status(
        contract_id="contracts",
        contract_version="2.0.0",
        dataset_id="orders",
        dataset_version="2024-04-01",
        status=status,
    )

    delete_queries = [
        query for query in spark.sql_queries if query.startswith("DELETE FROM analytics.status")
    ]
    assert delete_queries
    latest = delete_queries[-1]
    assert "dataset_id = 'orders'" in latest
    assert "dataset_version = '2024-04-01'" in latest


def test_save_status_deletes_existing_rows_for_path_target(tmp_path: Path) -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(spark, base_path=tmp_path)

    status = ValidationResult(status="ok")

    store.save_status(
        contract_id="contracts",
        contract_version="2.0.0",
        dataset_id="orders",
        dataset_version="2024-04-01",
        status=status,
    )

    delete_queries = [
        query for query in spark.sql_queries if query.startswith("DELETE FROM delta.`")
    ]
    assert delete_queries
    expected = str(tmp_path / "status")
    assert DeltaGovernanceStore._escape_identifier(expected) in delete_queries[-1]


def test_save_status_appends_metrics_to_derived_table_target() -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(
        spark,
        status_table="analytics.status",
        activity_table="analytics.activity",
        link_table="analytics.links",
    )
    store._now = lambda: "2024-04-05T12:00:00Z"  # type: ignore[assignment]

    status = ValidationResult(status="ok", metrics={"violations.total": 1})

    store.save_status(
        contract_id="contracts",
        contract_version="2.0.0",
        dataset_id="orders",
        dataset_version="2024-04-01",
        status=status,
    )

    metric_writes = [
        entry
        for entry in spark.records
        if entry["target"] == "analytics.status_metrics"
    ]
    assert metric_writes
    assert all(entry["type"] == "table" for entry in metric_writes)


def test_delta_store_normalises_string_metrics(tmp_path: Path) -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(spark, base_path=tmp_path)
    store._now = lambda: "2024-04-05T12:00:00Z"  # type: ignore[assignment]

    status = ValidationResult(status="ok", metrics={"row_count": "9", "note": "pass"})

    store.save_status(
        contract_id="contracts",
        contract_version="2.0.0",
        dataset_id="orders",
        dataset_version="2024-04-01",
        status=status,
    )

    metric_frames = [
        frame for frame in spark.dataframes if frame.get("schema") is store._METRIC_SCHEMA  # type: ignore[attr-defined]
    ]
    assert metric_frames
    records = metric_frames[-1]["data"]
    metric_map = {entry["metric_key"]: entry for entry in records}
    assert metric_map["row_count"]["metric_numeric_value"] == 9.0
    assert metric_map["row_count"]["metric_value"] == "9"
    assert metric_map["note"]["metric_value"] == "pass"


def test_load_metrics_filters_results(tmp_path: Path) -> None:
    spark = _StubSpark()
    store = DeltaGovernanceStore(spark, base_path=tmp_path, bootstrap_tables=False)

    rows = [
        {
            "dataset_id": "orders",
            "dataset_version": "2024-03-01",
            "contract_id": "contracts",
            "contract_version": "1.0.0",
            "status_recorded_at": "2024-03-02T00:00:00Z",
            "metric_key": "violations.total",
            "metric_value": "0",
            "metric_numeric_value": 0.0,
        },
        {
            "dataset_id": "orders",
            "dataset_version": "2024-03-02",
            "contract_id": "contracts",
            "contract_version": "1.0.0",
            "status_recorded_at": "2024-03-03T00:00:00Z",
            "metric_key": "summary",
            "metric_value": "{\"passed\": 10}",
            "metric_numeric_value": None,
        },
        {
            "dataset_id": "returns",
            "dataset_version": "2024-03-01",
            "contract_id": "contracts",
            "contract_version": "1.0.0",
            "status_recorded_at": "2024-03-02T00:00:00Z",
            "metric_key": "violations.total",
            "metric_value": "1",
            "metric_numeric_value": 1.0,
        },
    ]

    class _FakeRow(dict):
        def asDict(self, recursive: bool = False) -> dict[str, object]:
            return dict(self)

    class _FakeDataFrame:
        def __init__(self, data: list[dict[str, object]]) -> None:
            self._data = data

        def filter(self, _condition: object) -> "_FakeDataFrame":
            return self

        def orderBy(self, *_: object) -> "_FakeDataFrame":
            return self

        def collect(self) -> list[_FakeRow]:
            return [_FakeRow(row) for row in self._data]

    target_dataset_id = "orders"
    target_version = "2024-03-01"

    def _fake_read(**_: object) -> _FakeDataFrame:
        filtered = [
            row
            for row in rows
            if row["dataset_id"] == target_dataset_id and row["dataset_version"] == target_version
        ]
        return _FakeDataFrame(filtered)

    store._read = _fake_read  # type: ignore[assignment]

    from dc43_service_backends.governance.storage import delta as delta_module

    class _StubColumn:
        def __init__(self, name: str) -> None:
            self.name = name

        def __eq__(self, _: object) -> "_StubColumn":
            return self

        def __and__(self, _: object) -> "_StubColumn":
            return self

        def desc(self) -> "_StubColumn":  # pragma: no cover - behaviour ignored in stub
            return self

    stub_factory = lambda name: _StubColumn(name)

    original_col = delta_module.col
    delta_module.col = stub_factory  # type: ignore[assignment]

    load_metrics_globals = DeltaGovernanceStore.load_metrics.__globals__
    original_load_metrics_col = load_metrics_globals["col"]
    load_metrics_globals["col"] = stub_factory  # type: ignore[index]
    try:
        entries = store.load_metrics(
            dataset_id=target_dataset_id,
            dataset_version=target_version,
        )
    finally:
        delta_module.col = original_col  # type: ignore[assignment]
        load_metrics_globals["col"] = original_load_metrics_col

    assert len(entries) == 1
    entry = entries[0]
    assert entry["metric_key"] == "violations.total"
    assert entry["metric_numeric_value"] == 0.0



"""Core helpers shared across dc43 packages."""

from __future__ import annotations

from dc43_core import (  # noqa: F401 re-export
    BITOL_SCHEMA_URL,
    ODCS_REQUIRED,
    ODPS_REQUIRED,
    SemVer,
    as_odcs_dict,
    build_odcs,
    contract_identity,
    custom_properties_dict,
    ensure_version,
    field_map,
    fingerprint,
    list_properties,
    normalise_custom_properties,
    odcs_package_version,
    to_model,
    DataProductInputPort,
    DataProductOutputPort,
    OpenDataProduct,
    as_odps_product_dict,
    evolve_odps_to_draft,
    next_odps_draft_version,
    to_odps_model,
)

__all__ = [
    "BITOL_SCHEMA_URL",
    "ODCS_REQUIRED",
    "ODPS_REQUIRED",
    "SemVer",
    "as_odcs_dict",
    "build_odcs",
    "contract_identity",
    "custom_properties_dict",
    "ensure_version",
    "field_map",
    "fingerprint",
    "list_properties",
    "normalise_custom_properties",
    "odcs_package_version",
    "to_model",
    "DataProductInputPort",
    "DataProductOutputPort",
    "OpenDataProduct",
    "as_odps_product_dict",
    "evolve_odps_to_draft",
    "next_odps_draft_version",
    "to_odps_model",
]



from __future__ import annotations

from pathlib import Path

import pytest
import tomllib

from dc43_service_backends.config import (
    AuthConfig,
    ContractStoreConfig,
    DataProductStoreConfig,
    DataQualityBackendConfig,
    GovernanceConfig,
    GovernanceStoreConfig,
    ServiceBackendsConfig,
    UnityCatalogConfig,
    config_to_mapping,
    dumps,
    load_config,
)


def test_load_config_from_file(tmp_path: Path) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[contract_store]",
                f"root = '{tmp_path / 'contracts'}'",
                "", 
                "[data_product]",
                f"root = '{tmp_path / 'products'}'",
                "",
                "[data_quality]",
                "type = 'local'",
                "",
                "[auth]",
                "token = 'secret'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.contract_store.type == "filesystem"
    assert config.contract_store.root == tmp_path / "contracts"
    assert config.data_product_store.type == "memory"
    assert config.data_product_store.root == tmp_path / "products"
    assert config.data_quality.type == "local"
    assert config.data_quality.base_url is None
    assert config.data_quality.default_engine == "native"
    assert config.auth.token == "secret"
    assert config.governance.dataset_contract_link_builders == ()
    assert config.governance_store.type == "memory"


def test_load_config_env_overrides(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text("", encoding="utf-8")

    monkeypatch.setenv("DC43_SERVICE_BACKENDS_CONFIG", str(config_path))
    monkeypatch.setenv("DC43_CONTRACT_STORE", str(tmp_path / "override"))
    monkeypatch.setenv("DC43_BACKEND_TOKEN", "env-token")
    monkeypatch.setenv("DC43_DATA_PRODUCT_STORE", str(tmp_path / "dp"))
    monkeypatch.setenv("DC43_DATA_QUALITY_BACKEND_TYPE", "http")
    monkeypatch.setenv("DC43_DATA_QUALITY_BACKEND_URL", "https://quality.local")
    monkeypatch.setenv("DC43_DATA_QUALITY_BACKEND_TOKEN", "dq-token")
    monkeypatch.setenv("DC43_DATA_QUALITY_BACKEND_TOKEN_HEADER", "X-Api")
    monkeypatch.setenv("DC43_DATA_QUALITY_BACKEND_TOKEN_SCHEME", "")
    monkeypatch.setenv("DC43_GOVERNANCE_STORE_TYPE", "filesystem")
    monkeypatch.setenv("DC43_GOVERNANCE_STORE", str(tmp_path / "gov"))
    monkeypatch.setenv("DC43_DATA_QUALITY_DEFAULT_ENGINE", "soda")

    config = load_config()
    assert config.contract_store.type == "filesystem"
    assert config.contract_store.root == tmp_path / "override"
    assert config.data_product_store.root == tmp_path / "dp"
    assert config.data_quality.type == "http"
    assert config.data_quality.base_url == "https://quality.local"
    assert config.data_quality.token == "dq-token"
    assert config.data_quality.token_header == "X-Api"
    assert config.data_quality.token_scheme == ""
    assert config.data_quality.default_engine == "soda"
    assert config.auth.token == "env-token"
    assert config.unity_catalog.enabled is False
    assert config.governance.dataset_contract_link_builders == ()
    assert config.governance_store.type == "filesystem"
    assert config.governance_store.root == tmp_path / "gov"


def test_load_config_log_sql_flags(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[contract_store]",
                "type = 'sql'",
                "dsn = 'sqlite:///:memory:'",
                "log_sql = true",
                "",
                "[data_product]",
                "type = 'sql'",
                "dsn = 'sqlite:///:memory:'",
                "log_sql = false",
                "",
                "[governance_store]",
                "type = 'sql'",
                "dsn = 'sqlite:///:memory:'",
                "log_sql = false",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.contract_store.log_sql is True
    assert config.data_product_store.log_sql is False
    assert config.governance_store.log_sql is False

    monkeypatch.setenv("DC43_SERVICE_BACKENDS_CONFIG", str(config_path))
    monkeypatch.setenv("DC43_DATA_PRODUCT_STORE_LOG_SQL", "1")
    monkeypatch.setenv("DC43_GOVERNANCE_STORE_LOG_SQL", "true")

    env_config = load_config()
    assert env_config.contract_store.log_sql is True
    assert env_config.data_product_store.log_sql is True
    assert env_config.governance_store.log_sql is True


def test_governance_metrics_table_config(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[governance_store]",
                "type = 'sql'",
                "metrics_table = 'dq_status_metrics'",
                "status_table = 'dq_status'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.governance_store.metrics_table == "dq_status_metrics"

    monkeypatch.setenv("DC43_SERVICE_BACKENDS_CONFIG", str(config_path))
    monkeypatch.setenv("DC43_GOVERNANCE_METRICS_TABLE", "custom_metrics")

    env_config = load_config()
    assert env_config.governance_store.metrics_table == "custom_metrics"


def test_env_overrides_contract_store_type(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text("", encoding="utf-8")

    monkeypatch.setenv("DC43_SERVICE_BACKENDS_CONFIG", str(config_path))
    monkeypatch.setenv("DC43_CONTRACT_STORE_TYPE", "SQL")
    monkeypatch.setenv("DC43_CONTRACT_STORE_DSN", "sqlite:///example.db")

    config = load_config()
    assert config.contract_store.type == "sql"
    assert config.contract_store.dsn == "sqlite:///example.db"


def test_load_collibra_stub_config(tmp_path: Path) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[contract_store]",
                "type = 'collibra_stub'",
                "base_path = './stub-cache'",
                "default_status = 'Validated'",
                "status_filter = 'Validated'",
                "",
                "[contract_store.catalog.contract_a]",
                "data_product = 'dp-a'",
                "port = 'port-a'",
                "",
                "[contract_store.catalog.'contract-b']",
                "data_product = 'dp-b'",
                "port = 'port-b'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.contract_store.type == "collibra_stub"
    assert config.contract_store.base_path == Path("./stub-cache").expanduser()
    assert config.contract_store.default_status == "Validated"
    assert config.contract_store.status_filter == "Validated"
    assert config.contract_store.catalog == {
        "contract_a": ("dp-a", "port-a"),
        "contract-b": ("dp-b", "port-b"),
    }


def test_unity_catalog_tag_configuration(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[unity_catalog]",
                "enabled = true",
                "tags_enabled = true",
                "tags_sql_dsn = 'databricks://token@host'",
                "",
                "[unity_catalog.static_tags]",
                "owner = 'data-team'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.unity_catalog.tags_enabled is True
    assert config.unity_catalog.tags_sql_dsn == "databricks://token@host"
    assert config.unity_catalog.static_tags == {"owner": "data-team"}

    monkeypatch.setenv("DC43_SERVICE_BACKENDS_CONFIG", str(config_path))
    monkeypatch.setenv("DC43_UNITY_CATALOG_TAGS_ENABLED", "0")
    monkeypatch.setenv("DC43_UNITY_CATALOG_TAGS_SQL_DSN", "databricks://env@host")

    env_config = load_config()
    assert env_config.unity_catalog.tags_enabled is False
    assert env_config.unity_catalog.tags_sql_dsn == "databricks://env@host"


def test_delta_store_config(tmp_path: Path) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[contract_store]",
                "type = 'delta'",
                "table = 'governed.meta.contracts'",
                "",
                "[data_product]",
                "type = 'delta'",
                "table = 'governed.meta.data_products'",
                "",
                "[data_quality]",
                "type = 'http'",
                "base_url = 'https://observability.example.com'",
                "token = 'api-token'",
                "token_header = 'X-Token'",
                "token_scheme = ''",
                "",
                "[data_quality.headers]",
                "X-Org = 'governed'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.contract_store.type == "delta"
    assert config.contract_store.table == "governed.meta.contracts"
    assert config.data_product_store.type == "delta"
    assert config.data_product_store.table == "governed.meta.data_products"
    assert config.data_quality.type == "http"
    assert config.data_quality.base_url == "https://observability.example.com"
    assert config.data_quality.token == "api-token"
    assert config.data_quality.token_header == "X-Token"
    assert config.data_quality.token_scheme == ""
    assert config.data_quality.headers == {"X-Org": "governed"}


def test_data_quality_engines_config(tmp_path: Path) -> None:
    suite_path = tmp_path / "suite.json"
    suite_path.write_text("{}", encoding="utf-8")
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[data_quality]",
                "default_engine = 'great_expectations'",
                "",
                "[data_quality.engines.native]",
                "type = 'native'",
                "strict_types = false",
                "",
                "[data_quality.engines.great_expectations]",
                f"suite_path = '{suite_path}'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.data_quality.default_engine == "great_expectations"
    assert "native" in config.data_quality.engines
    native_cfg = config.data_quality.engines["native"]
    assert native_cfg["type"] == "native"
    assert native_cfg["strict_types"] is False
    ge_cfg = config.data_quality.engines["great_expectations"]
    assert ge_cfg["suite_path"] == str(suite_path)


def test_load_collibra_http_config(tmp_path: Path) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[contract_store]",
                "type = 'collibra_http'",
                "base_url = 'https://collibra.example.com'",
                "token = 'api-token'",
                "timeout = 5.5",
                "contracts_endpoint_template = '/custom/{data_product}/{port}'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.contract_store.type == "collibra_http"
    assert config.contract_store.base_url == "https://collibra.example.com"
    assert config.contract_store.token == "api-token"
    assert config.contract_store.timeout == 5.5
    assert config.contract_store.contracts_endpoint_template == "/custom/{data_product}/{port}"


def test_unity_catalog_config_section(tmp_path: Path) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text(
        "\n".join(
            [
                "[unity_catalog]",
                "enabled = true",
                "dataset_prefix = 'table:'",
                "workspace_profile = 'prod'",
                "workspace_url = 'https://adb.example.com'",
                "workspace_token = 'token-123'",
                "sql_dsn = 'databricks://token:abc@adb.example.com?http_path=/sql/1.0/warehouses/123'",
                "",
                "[unity_catalog.static_properties]",
                "owner = 'governance'",
                "",
                "[governance]",
                "dataset_contract_link_builders = [",
                "  'example.module:builder',",
                "]",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_config(config_path)
    assert config.unity_catalog.enabled is True
    assert config.unity_catalog.dataset_prefix == "table:"
    assert config.unity_catalog.workspace_profile == "prod"
    assert config.unity_catalog.workspace_url == "https://adb.example.com"
    assert config.unity_catalog.workspace_token == "token-123"
    assert (
        config.unity_catalog.sql_dsn
        == "databricks://token:abc@adb.example.com?http_path=/sql/1.0/warehouses/123"
    )
    assert config.unity_catalog.static_properties == {"owner": "governance"}
    assert config.governance.dataset_contract_link_builders == (
        "example.module:builder",
    )


def test_unity_catalog_env_overrides(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path = tmp_path / "backends.toml"
    config_path.write_text("", encoding="utf-8")

    monkeypatch.setenv("DC43_SERVICE_BACKENDS_CONFIG", str(config_path))
    monkeypatch.setenv("DC43_UNITY_CATALOG_ENABLED", "yes")
    monkeypatch.setenv("DC43_UNITY_CATALOG_PREFIX", "cat:")
    monkeypatch.setenv("DATABRICKS_CONFIG_PROFILE", "unity-prod")
    monkeypatch.setenv("DATABRICKS_HOST", "https://adb.example.com")
    monkeypatch.setenv("DATABRICKS_TOKEN", "env-token")
    monkeypatch.setenv(
        "DC43_UNITY_CATALOG_SQL_DSN",
        "databricks://token:env@adb.example.com?http_path=/sql/warehouses/abc",
    )
    monkeypatch.setenv(
        "DC43_GOVERNANCE_LINK_BUILDERS",
        "custom.module:builder, other.module.hooks:make",
    )

    config = load_config()
    assert config.unity_catalog.enabled is True
    assert config.unity_catalog.dataset_prefix == "cat:"
    assert config.unity_catalog.workspace_profile == "unity-prod"
    assert config.unity_catalog.workspace_url == "https://adb.example.com"
    assert config.unity_catalog.workspace_token == "env-token"
    assert (
        config.unity_catalog.sql_dsn
        == "databricks://token:env@adb.example.com?http_path=/sql/warehouses/abc"
    )
    assert config.governance.dataset_contract_link_builders == (
        "custom.module:builder",
        "other.module.hooks:make",
    )


def test_dumps_matches_mapping_including_workspace_url() -> None:
    config = ServiceBackendsConfig(
        contract_store=ContractStoreConfig(type="delta", table="governed.contracts"),
        data_quality=DataQualityBackendConfig(
            type="http",
            base_url="https://quality.example.com",
            token="dq-token",
            headers={"X-Env": "prod"},
        ),
        unity_catalog=UnityCatalogConfig(
            enabled=True,
            workspace_url="https://adb.example.com",
            workspace_profile="prod",
            workspace_token="token-123",
            dataset_prefix="table:",
            sql_dsn="databricks://token:abc@adb.example.com?http_path=/sql/warehouses/abc",
            static_properties={"catalog": "main", "schema": "contracts"},
        ),
        auth=AuthConfig(token="auth-token"),
    )

    toml_text = dumps(config)
    parsed = tomllib.loads(toml_text)

    assert parsed == config_to_mapping(config)
    assert parsed["unity_catalog"]["workspace_url"] == "https://adb.example.com"


def test_dumps_handles_missing_tomlkit(monkeypatch: pytest.MonkeyPatch) -> None:
    from dc43_service_backends import config as backends_config

    service_config = ServiceBackendsConfig(
        contract_store=ContractStoreConfig(type="filesystem"),
        data_product_store=DataProductStoreConfig(type="memory"),
        auth=AuthConfig(token="shared"),
    )

    original = backends_config.tomlkit
    monkeypatch.setattr(backends_config, "tomlkit", None)
    try:
        toml_text = backends_config.dumps(service_config)
    finally:
        monkeypatch.setattr(backends_config, "tomlkit", original)

    parsed = tomllib.loads(toml_text)
    assert parsed == config_to_mapping(service_config)


def test_config_to_mapping_covers_all_sections() -> None:
    config = ServiceBackendsConfig(
        contract_store=ContractStoreConfig(
            type="collibra_http",
            root=Path("/var/contracts"),
            base_path=Path("/delta/contracts"),
            table="governance.contracts",
            base_url="https://contracts.example.com",
            dsn="postgresql+psycopg://user:pass@host/contracts",
            schema="governance",
            token="contract-token",
            timeout=22.5,
            contracts_endpoint_template="/custom/contracts/{data_product}/{port}",
            default_status="Validated",
            status_filter="Validated",
            catalog={"product-quality": ("dp-quality", "gold")},
        ),
        data_product_store=DataProductStoreConfig(
            type="collibra_http",
            root=Path("/var/products"),
            base_path=Path("/delta/products"),
            table="governance.products",
            dsn="postgresql+psycopg://user:pass@host/products",
            schema="governance",
            base_url="https://products.example.com",
            catalog="governance-products",
        ),
        data_quality=DataQualityBackendConfig(
            type="http",
            base_url="https://dq.example.com",
            token="dq-token",
            token_header="X-Token",
            token_scheme="Token",
            headers={"X-Team": "quality"},
            default_engine="soda",
            engines={
                "native": {"strict_types": False},
                "great_expectations": {"suite_path": "/tmp/suite.json"},
            },
        ),
        auth=AuthConfig(token="shared-token"),
        unity_catalog=UnityCatalogConfig(
            enabled=True,
            dataset_prefix="table:",
            workspace_profile="prod",
            workspace_url="https://adb.example.com",
            workspace_token="uc-token",
            sql_dsn="databricks://token:abc@adb.example.com?http_path=/sql/warehouses/abc",
            static_properties={"catalog": "main", "schema": "governance"},
        ),
        governance=GovernanceConfig(
            dataset_contract_link_builders=("custom.module:builder",),
        ),
        governance_store=GovernanceStoreConfig(
            type="http",
            root=Path("/var/governance"),
            base_path=Path("/delta/governance"),
            table="governance.status",
            status_table="governance.status_history",
            activity_table="governance.activity",
            link_table="governance.links",
            dsn="postgresql+psycopg://user:pass@host/governance",
            schema="governance",
            base_url="https://governance.example.com",
            token="governance-token",
            token_header="X-Gov-Token",
            token_scheme="Token",
            timeout=45.5,
            headers={"X-Env": "prod"},
        ),
    )

    mapping = config_to_mapping(config)

    assert mapping["contract_store"] == {
        "type": "collibra_http",
        "root": str(Path("/var/contracts")),
        "base_path": str(Path("/delta/contracts")),
        "table": "governance.contracts",
        "base_url": "https://contracts.example.com",
        "dsn": "postgresql+psycopg://user:pass@host/contracts",
        "schema": "governance",
        "token": "contract-token",
        "timeout": 22.5,
        "contracts_endpoint_template": "/custom/contracts/{data_product}/{port}",
        "default_status": "Validated",
        "status_filter": "Validated",
        "catalog": {
            "product-quality": {
                "data_product": "dp-quality",
                "port": "gold",
            }
        },
    }
    assert mapping["data_product"] == {
        "type": "collibra_http",
        "root": str(Path("/var/products")),
        "base_path": str(Path("/delta/products")),
        "table": "governance.products",
        "dsn": "postgresql+psycopg://user:pass@host/products",
        "schema": "governance",
        "base_url": "https://products.example.com",
        "catalog": "governance-products",
    }
    assert mapping["data_quality"] == {
        "type": "http",
        "base_url": "https://dq.example.com",
        "token": "dq-token",
        "token_header": "X-Token",
        "token_scheme": "Token",
        "headers": {"X-Team": "quality"},
        "default_engine": "soda",
        "engines": {
            "native": {"strict_types": False},
            "great_expectations": {"suite_path": "/tmp/suite.json"},
        },
    }
    assert mapping["auth"] == {"token": "shared-token"}
    assert mapping["unity_catalog"] == {
        "enabled": True,
        "dataset_prefix": "table:",
        "workspace_profile": "prod",
        "workspace_url": "https://adb.example.com",
        "workspace_token": "uc-token",
        "sql_dsn": "databricks://token:abc@adb.example.com?http_path=/sql/warehouses/abc",
        "static_properties": {"catalog": "main", "schema": "governance"},
    }
    assert mapping["governance"] == {
        "dataset_contract_link_builders": ["custom.module:builder"],
    }
    assert mapping["governance_store"] == {
        "type": "http",
        "root": str(Path("/var/governance")),
        "base_path": str(Path("/delta/governance")),
        "table": "governance.status",
        "status_table": "governance.status_history",
        "activity_table": "governance.activity",
        "link_table": "governance.links",
        "dsn": "postgresql+psycopg://user:pass@host/governance",
        "schema": "governance",
        "base_url": "https://governance.example.com",
        "token": "governance-token",
        "token_header": "X-Gov-Token",
        "token_scheme": "Token",
        "timeout": 45.5,
        "headers": {"X-Env": "prod"},
    }



"""Reference contract store implementations."""

from .filesystem import FSContractStore
from .delta import DeltaContractStore

try:  # pragma: no cover - optional dependency
    from .sql import SQLContractStore
except ModuleNotFoundError:  # pragma: no cover
    SQLContractStore = None  # type: ignore
from .collibra import (
    CollibraContractAdapter,
    CollibraContractGateway,
    CollibraContractStore,
    ContractSummary,
    HttpCollibraContractAdapter,
    HttpCollibraContractGateway,
    StubCollibraContractAdapter,
    StubCollibraContractGateway,
)

__all__ = [
    "CollibraContractAdapter",
    "CollibraContractGateway",
    "CollibraContractStore",
    "ContractSummary",
    "DeltaContractStore",
    "FSContractStore",
    "HttpCollibraContractAdapter",
    "HttpCollibraContractGateway",
    "StubCollibraContractAdapter",
    "StubCollibraContractGateway",
]

if SQLContractStore is not None:
    __all__.append("SQLContractStore")



"""Service backend implementations for dc43."""

from __future__ import annotations

from importlib import import_module
from typing import Any

__all__ = [
    "auth",
    "config",
    "bootstrap",
    "contracts",
    "data_products",
    "data_quality",
    "governance",
    "server",
    "web",
]


def __getattr__(name: str) -> Any:
    if name in __all__:
        module = import_module(f".{name}", __name__)
        globals()[name] = module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__() -> list[str]:
    return sorted(set(globals()) | set(__all__))



from types import SimpleNamespace

import pytest

from dc43_service_backends.data_quality.backend.manager import DataQualityManager
from dc43_service_clients.data_quality import ObservationPayload, ValidationResult


class _StubEngine:
    def __init__(self, name: str) -> None:
        self.name = name
        self.calls: list[tuple[str, object]] = []

    def evaluate(self, contract: object, payload: ObservationPayload) -> ValidationResult:
        self.calls.append(("evaluate", contract))
        return ValidationResult(status=self.name, details={"engine": self.name})

    def describe_expectations(self, contract: object) -> list[dict[str, object]]:
        self.calls.append(("describe", contract))
        return [{"engine": self.name}]


def _contract_with_engine(engine: str | None = None) -> object:
    metadata = {"quality_engine": engine} if engine else {}
    schema_object = SimpleNamespace(quality=[])
    return SimpleNamespace(metadata=metadata, schema_=[schema_object])


def test_manager_uses_default_engine_when_contract_omits_override() -> None:
    engine = _StubEngine("soda")
    manager = DataQualityManager(default_engine="soda", engines={"soda": engine})

    contract = _contract_with_engine()
    payload = ObservationPayload(metrics={})
    result = manager.evaluate(contract, payload)

    assert result.status == "soda"
    assert engine.calls and engine.calls[0][0] == "evaluate"


def test_manager_prefers_contract_metadata_engine() -> None:
    engine = _StubEngine("great_expectations")
    manager = DataQualityManager(
        default_engine="native",
        engines={"great_expectations": engine},
    )

    contract = _contract_with_engine("great_expectations")
    payload = ObservationPayload(metrics={"great_expectations": {"status": "ok"}})
    result = manager.evaluate(contract, payload)

    assert result.status == "great_expectations"
    assert engine.calls and engine.calls[0][0] == "evaluate"


def test_manager_raises_for_unknown_engine() -> None:
    manager = DataQualityManager(default_engine="custom")
    contract = _contract_with_engine("nonexistent")
    with pytest.raises(RuntimeError):
        manager.evaluate(contract, ObservationPayload(metrics={}))



import pytest

from open_data_contract_standard.model import (
    Description,
    OpenDataContractStandard as ODCSModel,
    SchemaObject,
    SchemaProperty,
)

from dc43_service_backends.core.odps import (
    DataProductInputPort,
    DataProductOutputPort,
    OpenDataProductStandard,
    as_odps_dict,
    evolve_to_draft,
    next_draft_version,
)


def test_next_draft_version_uniqueness() -> None:
    existing = ["0.2.0-draft", "0.2.0-draft.2"]
    version = next_draft_version(
        current_version="0.1.0",
        existing_versions=existing,
        bump="minor",
    )
    assert version.startswith("0.2.0-draft")
    assert version not in existing


def test_evolve_to_draft_sets_status_and_version() -> None:
    product = OpenDataProductStandard(
        id="dp.orders",
        status="active",
        version="1.0.0",
    )
    evolve_to_draft(product, existing_versions=["1.0.0"], bump="minor")
    assert product.status == "draft"
    assert product.version is not None
    assert product.version.startswith("1.1.0-draft")


def test_ensure_input_port_replaces_existing() -> None:
    product = OpenDataProductStandard(id="dp.sales", status="draft")
    port_a = DataProductInputPort(name="orders", version="1.0.0", contract_id="orders")
    product.ensure_input_port(port_a)
    port_b = DataProductInputPort(name="orders", version="1.1.0", contract_id="orders")
    changed = product.ensure_input_port(port_b)
    assert changed is True
    assert product.input_ports[0].version == "1.1.0"


def test_ensure_output_port_idempotent() -> None:
    product = OpenDataProductStandard(id="dp.sales", status="draft")
    port = DataProductOutputPort(name="report", version="2.0.0", contract_id="report")
    first = product.ensure_output_port(port)
    second = product.ensure_output_port(port)
    assert first is True
    assert second is False


def test_as_odps_dict_rejects_contract() -> None:
    contract = ODCSModel(
        version="1.0.0",
        kind="DataContract",
        apiVersion="3.0.2",
        id="test.orders",
        name="Orders",
        description=Description(usage="Orders facts"),
        schema=[
            SchemaObject(
                name="orders",
                properties=[
                    SchemaProperty(name="order_id", physicalType="bigint", required=True)
                ],
            )
        ],
    )

    with pytest.raises(TypeError, match="OpenDataContractStandard"):
        as_odps_dict(contract)



"""Helpers for normalising metrics stored with validation results."""

from __future__ import annotations

import json
from numbers import Number
from typing import Any, Mapping, Tuple

from dc43_service_clients.data_quality import ValidationResult


def extract_metrics(status: ValidationResult | None) -> dict[str, Any]:
    """Return a serialisable metrics mapping for ``status``.

    Some validation providers only attach metric observations inside the
    :pyattr:`ValidationResult.details` payload instead of the ``metrics``
    attribute. The governance stores expect explicit values to populate the
    ``dq_metrics`` tables, so we merge both sources to avoid dropping data.
    """

    if status is None:
        return {}

    metrics: dict[str, Any] = {}
    details = status.details
    detail_metrics = details.get("metrics") if isinstance(details, Mapping) else None
    if isinstance(detail_metrics, Mapping):
        metrics.update(detail_metrics)
    metrics.update(status.metrics or {})
    return metrics


def normalise_metric_value(value: Any) -> Tuple[str | None, float | None]:
    """Return storage-friendly metric payloads plus numeric hints.

    SQL, Delta, and filesystem stores persist metric observations so the UI can
    display historical charts. The persistence layer stores values as text, but
    callers often provide numbers (or numeric strings) that should populate the
    ``metric_numeric_value`` field to simplify filtering. This helper mirrors the
    UI's coercion logic by returning the serialised value alongside a
    floating-point representation whenever the input resembles a number.
    """

    if value is None:
        return None, None
    if isinstance(value, Number):
        return str(value), float(value)
    if isinstance(value, str):
        stripped = value.strip()
        if stripped:
            try:
                numeric = float(stripped)
            except ValueError:
                numeric = None
        else:
            numeric = None
        return value, numeric
    try:
        serialised = json.dumps(value)
    except TypeError:
        serialised = str(value)
    return serialised, None


__all__ = ["extract_metrics", "normalise_metric_value"]




"""Helpers to project expectation specs into serialisable plans."""

from __future__ import annotations

from typing import Any, Dict, Iterable, List, Mapping

from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

from .engine import ExpectationSpec, expectation_specs


def _sql_literal(value: Any) -> str:
    if isinstance(value, str):
        escaped = value.replace("'", "\\'")
        return f"'{escaped}'"
    if value is None:
        return "NULL"
    return str(value)


def sql_predicate(spec: ExpectationSpec) -> str | None:
    """Return a Spark SQL predicate for the provided expectation spec."""

    column = spec.column
    if not column:
        return None
    col_ref = f"`{column.replace('`', '')}`"
    if spec.rule in {"not_null", "required"}:
        return f"{col_ref} IS NOT NULL"
    if spec.rule == "gt":
        return f"{col_ref} > {_sql_literal(spec.params.get('threshold'))}"
    if spec.rule == "ge":
        return f"{col_ref} >= {_sql_literal(spec.params.get('threshold'))}"
    if spec.rule == "lt":
        return f"{col_ref} < {_sql_literal(spec.params.get('threshold'))}"
    if spec.rule == "le":
        return f"{col_ref} <= {_sql_literal(spec.params.get('threshold'))}"
    if spec.rule == "enum":
        values = spec.params.get("values") or []
        if not isinstance(values, (list, tuple, set)):
            return None
        literals = ", ".join(_sql_literal(v) for v in values)
        return f"{col_ref} IN ({literals})" if literals else None
    if spec.rule == "regex":
        pattern = spec.params.get("pattern")
        if pattern is None:
            return None
        pattern_str = str(pattern).replace("'", "\\'")
        return f"{col_ref} RLIKE '{pattern_str}'"
    return None


def expectation_plan(contract: OpenDataContractStandard) -> List[Dict[str, Any]]:
    """Return serialisable expectation descriptors derived from ``contract``."""

    plan: List[Dict[str, Any]] = []
    for spec in expectation_specs(contract):
        entry: Dict[str, Any] = {
            "key": spec.key,
            "rule": spec.rule,
            "column": spec.column,
            "optional": bool(spec.optional),
        }
        if spec.params:
            entry["params"] = dict(spec.params)
        predicate = sql_predicate(spec)
        if predicate:
            entry["predicate"] = predicate
        plan.append(entry)
    return plan


def expectation_predicates_from_plan(
    plan: Iterable[Mapping[str, Any]]
) -> Dict[str, str]:
    """Return ``expectation -> predicate`` from a plan when available."""

    mapping: Dict[str, str] = {}
    for item in plan:
        key = item.get("key")
        predicate = item.get("predicate")
        if isinstance(key, str) and isinstance(predicate, str):
            mapping[key] = predicate
    return mapping


__all__ = [
    "expectation_plan",
    "expectation_predicates_from_plan",
    "sql_predicate",
]



from __future__ import annotations

"""Filesystem-based contract store (DBFS/UC Volumes/local).

Stores ODCS documents under ``{base}/{contract_id}/{version}.json``.
"""

import logging
import os
from typing import List

from .interface import ContractStore
from dc43_service_backends.core.odcs import as_odcs_dict, ensure_version, contract_identity, to_model
from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

logger = logging.getLogger(__name__)


class FSContractStore(ContractStore):
    """Filesystem-based store (works with DBFS, UC Volumes, local FS)."""

    def __init__(self, base_path: str):
        self.base_path = base_path.rstrip("/")

    def _dir(self, contract_id: str) -> str:
        return os.path.join(self.base_path, contract_id)

    def _path(self, contract_id: str, version: str) -> str:
        return os.path.join(self._dir(contract_id), f"{version}.json")

    def put(self, contract: OpenDataContractStandard) -> None:
        """Write an ODCS document model to the filesystem."""
        ensure_version(contract)
        odcs = as_odcs_dict(contract)
        cid, ver = contract_identity(contract)
        d = self._dir(cid)
        os.makedirs(d, exist_ok=True)
        p = self._path(cid, ver)
        logger.info("Storing contract %s:%s at %s", cid, ver, p)
        with open(p, "w", encoding="utf-8") as f:
            import json

            f.write(json.dumps(odcs, indent=2, sort_keys=True))

    def get(self, contract_id: str, version: str) -> OpenDataContractStandard:
        """Read an ODCS document as a model from the filesystem."""
        p = self._path(contract_id, version)
        logger.info("Loading contract %s:%s from %s", contract_id, version, p)
        with open(p, "r", encoding="utf-8") as f:
            import json
            return to_model(json.loads(f.read()))

    def list_contracts(self) -> List[str]:
        """List all contract identifiers stored under the base path."""
        if not os.path.isdir(self.base_path):
            return []
        ids: List[str] = []
        for name in os.listdir(self.base_path):
            d = os.path.join(self.base_path, name)
            if os.path.isdir(d) and any(f.endswith(".json") for f in os.listdir(d)):
                ids.append(name)
        return ids

    def list_versions(self, contract_id: str) -> List[str]:
        """List versions present on disk for ``contract_id``."""
        d = self._dir(contract_id)
        if not os.path.isdir(d):
            return []
        versions: List[str] = []
        for name in os.listdir(d):
            if name.endswith(".json"):
                versions.append(name[:-5])
        return versions


__all__ = ["FSContractStore"]



"""Pluggable execution engines for the data-quality manager."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Mapping, Optional, Protocol, Sequence

from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

from dc43_service_backends.data_quality.backend.engine import (
    ValidationResult,
    evaluate_contract,
)
from dc43_service_backends.data_quality.backend.predicates import (
    expectation_plan,
    expectation_predicates_from_plan,
)
from dc43_service_clients.data_quality import ObservationPayload

try:  # pragma: no cover - optional dependency
    import yaml  # type: ignore
except ModuleNotFoundError:  # pragma: no cover - optional dependency
    yaml = None  # type: ignore[assignment]


class DataQualityExecutionEngine(Protocol):
    """Execution contract implemented by concrete data-quality engines."""

    def evaluate(
        self,
        contract: OpenDataContractStandard,
        payload: ObservationPayload,
    ) -> ValidationResult:
        """Return the validation outcome for ``contract`` and ``payload``."""

    def describe_expectations(
        self,
        contract: OpenDataContractStandard,
    ) -> Sequence[Mapping[str, object]]:
        """Describe expectations that the engine will enforce."""


class NativeDataQualityEngine(DataQualityExecutionEngine):
    """Adapter around the built-in contract evaluation engine."""

    def __init__(
        self,
        *,
        strict_types: bool = True,
        allow_extra_columns: bool = True,
        expectation_severity: str = "error",
    ) -> None:
        self._strict_types = strict_types
        self._allow_extra_columns = allow_extra_columns
        self._expectation_severity = expectation_severity

    def evaluate(
        self,
        contract: OpenDataContractStandard,
        payload: ObservationPayload,
    ) -> ValidationResult:
        result = evaluate_contract(
            contract,
            schema=payload.schema,
            metrics=payload.metrics,
            strict_types=self._strict_types,
            allow_extra_columns=self._allow_extra_columns,
            expectation_severity=self._expectation_severity,  # type: ignore[arg-type]
        )
        plan = expectation_plan(contract)
        updates: dict[str, object] = {}
        if plan:
            updates["expectation_plan"] = plan
            predicates = expectation_predicates_from_plan(plan)
            if predicates:
                updates["expectation_predicates"] = predicates
        if updates:
            result.merge_details(updates)
        return result

    def describe_expectations(
        self,
        contract: OpenDataContractStandard,
    ) -> Sequence[Mapping[str, object]]:
        return expectation_plan(contract)


class _SuiteLoader:
    """Utility helper to load expectation suites for external engines."""

    def __init__(self, path: str | Path | None) -> None:
        self._path = Path(path).expanduser() if path else None

    def load(self) -> Sequence[Mapping[str, object]]:
        if self._path is None or not self._path.exists():
            return []
        try:
            data = self._path.read_text("utf-8")
        except OSError:
            return []
        if self._path.suffix.lower() in {".json", ".ge"}:
            try:
                payload = json.loads(data)
            except json.JSONDecodeError:
                return []
        elif self._path.suffix.lower() in {".yaml", ".yml"} and yaml is not None:
            try:
                payload = yaml.safe_load(data)  # type: ignore[assignment]
            except Exception:
                return []
        else:
            return []
        if isinstance(payload, Mapping):
            return [dict(payload)]
        if isinstance(payload, list):
            return [item for item in payload if isinstance(item, Mapping)]
        return []


class MetricsDrivenEngine(DataQualityExecutionEngine):
    """Base class for engines that interpret metrics emitted by external tools."""

    def __init__(self, *, metrics_key: str, suite_path: str | Path | None = None) -> None:
        self._metrics_key = metrics_key
        self._suite_loader = _SuiteLoader(suite_path)

    def _metrics_summary(self, payload: ObservationPayload) -> Mapping[str, object] | None:
        metrics = payload.metrics or {}
        summary = metrics.get(self._metrics_key)
        if isinstance(summary, Mapping):
            return summary
        return None

    def describe_expectations(
        self,
        contract: OpenDataContractStandard,
    ) -> Sequence[Mapping[str, object]]:
        suite = self._suite_loader.load()
        if suite:
            return suite
        return expectation_plan(contract)


class GreatExpectationsEngine(MetricsDrivenEngine):
    """Translate Great Expectations run summaries into validation results."""

    def __init__(
        self,
        *,
        metrics_key: str = "great_expectations",
        suite_path: str | Path | None = None,
    ) -> None:
        super().__init__(metrics_key=metrics_key, suite_path=suite_path)

    def evaluate(
        self,
        contract: OpenDataContractStandard,
        payload: ObservationPayload,
    ) -> ValidationResult:
        summary = self._metrics_summary(payload)
        if not summary:
            return ValidationResult(
                status="unknown",
                reason="missing-great-expectations-summary",
                details={"engine": "great_expectations", "metrics_key": self._metrics_key},
            )
        status = str(summary.get("status") or ("ok" if summary.get("success") else "block"))
        if status not in {"ok", "warn", "block", "unknown"}:
            status = "ok" if summary.get("success") else "block"
        reason = summary.get("exception_info") or summary.get("reason")
        if isinstance(reason, Mapping):
            reason = reason.get("exception_message")
        return ValidationResult(
            status=status,
            reason=str(reason) if reason else None,
            details=dict(summary),
        )


class SodaEngine(MetricsDrivenEngine):
    """Interpret Soda scan outcomes from observation metrics."""

    def __init__(
        self,
        *,
        metrics_key: str = "soda",
        checks_path: str | Path | None = None,
    ) -> None:
        super().__init__(metrics_key=metrics_key, suite_path=checks_path)

    def evaluate(
        self,
        contract: OpenDataContractStandard,
        payload: ObservationPayload,
    ) -> ValidationResult:
        summary = self._metrics_summary(payload)
        if not summary:
            return ValidationResult(
                status="unknown",
                reason="missing-soda-summary",
                details={"engine": "soda", "metrics_key": self._metrics_key},
            )
        outcome = str(summary.get("status") or summary.get("outcome") or "warn")
        if outcome not in {"ok", "warn", "block", "unknown"}:
            outcome = "ok" if str(summary.get("status", "")).lower() in {"passed", "pass"} else "block"
        reason = summary.get("reason") or summary.get("error")
        return ValidationResult(
            status=outcome,
            reason=str(reason) if reason else None,
            details=dict(summary),
        )


__all__ = [
    "DataQualityExecutionEngine",
    "NativeDataQualityEngine",
    "GreatExpectationsEngine",
    "SodaEngine",
]



"""Data product service backends."""

from .backend import (
    CollibraDataProductAdapter,
    CollibraDataProductServiceBackend,
    DataProductRegistrationResult,
    DataProductServiceBackend,
    DeltaDataProductServiceBackend,
    FilesystemDataProductServiceBackend,
    HttpCollibraDataProductAdapter,
    LocalDataProductServiceBackend,
    StubCollibraDataProductAdapter,
)
from .backend.interface import DataProductListing

__all__ = [
    "CollibraDataProductAdapter",
    "CollibraDataProductServiceBackend",
    "DataProductRegistrationResult",
    "DataProductListing",
    "DataProductServiceBackend",
    "DeltaDataProductServiceBackend",
    "FilesystemDataProductServiceBackend",
    "HttpCollibraDataProductAdapter",
    "LocalDataProductServiceBackend",
    "StubCollibraDataProductAdapter",
]




"""Compatibility shim for contract drafting helpers."""

from __future__ import annotations

from typing import Any

__all__ = ["draft_from_observations", "draft_from_validation_result"]


def draft_from_observations(*args: Any, **kwargs: Any) -> Any:
    from .backend.drafting import draft_from_observations as _impl

    return _impl(*args, **kwargs)


def draft_from_validation_result(*args: Any, **kwargs: Any) -> Any:
    from .backend.drafting import draft_from_validation_result as _impl

    return _impl(*args, **kwargs)



from __future__ import annotations

from typing import Mapping, Sequence

import pytest
from open_data_contract_standard.model import (  # type: ignore
    OpenDataContractStandard,
    SchemaObject,
    SchemaProperty,
    Server,
)

from dc43_service_backends.contracts.backend.local import LocalContractServiceBackend
from dc43_service_backends.contracts.backend.stores.filesystem import FSContractStore
from dc43_service_backends.data_quality import LocalDataQualityServiceBackend
from dc43_service_backends.governance.backend.local import LocalGovernanceServiceBackend
from dc43_service_clients.data_quality import ObservationPayload, ValidationResult
from dc43_service_clients.data_products.models import (
    DataProductInputBinding,
    DataProductOutputBinding,
)
from dc43_service_clients.governance.models import (
    ContractReference,
    GovernancePolicy,
    GovernanceReadContext,
    GovernanceWriteContext,
)
from dc43_service_clients.governance.lineage import decode_lineage_event
from dc43_service_clients.odps import (
    DataProductInputPort,
    DataProductOutputPort,
    OpenDataProductStandard,
)
from dc43_service_clients.testing.backends import LocalDataProductServiceBackend


class RecordingDataProductBackend(LocalDataProductServiceBackend):
    """Capture registration calls for assertions."""

    def __init__(self) -> None:
        super().__init__()
        self.last_input_call: Mapping[str, object] | None = None
        self.last_output_call: Mapping[str, object] | None = None

    def register_input_port(self, **kwargs):  # type: ignore[override]
        self.last_input_call = dict(kwargs)
        return super().register_input_port(**kwargs)

    def register_output_port(self, **kwargs):  # type: ignore[override]
        self.last_output_call = dict(kwargs)
        return super().register_output_port(**kwargs)


def _sample_contract(version: str = "1.0.0") -> OpenDataContractStandard:
    return OpenDataContractStandard(
        version=version,
        kind="DatasetContract",
        apiVersion="3.0.2",
        id="sales.orders",
        name="Sales Orders",
        schema=[
            SchemaObject(
                name="orders",
                properties=[
                    SchemaProperty(
                        name="order_id",
                        physicalType="integer",
                        required=True,
                    ),
                    SchemaProperty(
                        name="order_ts",
                        physicalType="string",
                    ),
                ],
            )
        ],
        servers=[
            Server(server="s3", type="s3", path="datalake/orders", format="delta")
        ],
    )


@pytest.fixture()
def governance_fixture(tmp_path):
    contract = _sample_contract()
    store = FSContractStore(str(tmp_path / "contracts"))
    contract_backend = LocalContractServiceBackend(store)
    contract_backend.put(contract)

    dq_backend = LocalDataQualityServiceBackend()
    data_product_backend = RecordingDataProductBackend()
    data_product_backend.register_input_port(
        data_product_id="dp.analytics",
        port=DataProductInputPort(
            name="orders", version=contract.version, contract_id=contract.id
        ),
    )
    data_product_backend.register_output_port(
        data_product_id="dp.analytics",
        port=DataProductOutputPort(
            name="primary", version=contract.version, contract_id=contract.id
        ),
    )

    product = data_product_backend.latest("dp.analytics")
    if product is not None:
        product.status = "active"
        if product.version and product.version.endswith("-draft"):
            product.version = product.version[: -len("-draft")]
        data_product_backend.put(product)

    backend = LocalGovernanceServiceBackend(
        contract_client=contract_backend,
        dq_client=dq_backend,
        data_product_client=data_product_backend,
        draft_store=store,
    )
    return backend, data_product_backend, contract


def test_resolve_read_context_uses_data_product_binding(governance_fixture):
    backend, data_product_backend, contract = governance_fixture

    context = GovernanceReadContext(
        input_binding=DataProductInputBinding(
            data_product="dp.analytics",
            port_name="orders",
        ),
        dataset_id="analytics.orders",
        policy=GovernancePolicy(draft_on_violation=True),
    )

    plan = backend.resolve_read_context(context=context)
    assert plan.contract_id == contract.id
    assert plan.dataset_id == "analytics.orders"
    assert plan.input_binding is not None

    assessment = backend.evaluate_read_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )
    backend.register_read_activity(plan=plan, assessment=assessment)

    assert data_product_backend.last_input_call is not None
    assert data_product_backend.last_input_call["data_product_id"] == "dp.analytics"


def test_register_read_activity_respects_binding_version(governance_fixture):
    backend, data_product_backend, contract = governance_fixture

    product_v1 = OpenDataProductStandard(
        id="Sales.KPIs",
        status="draft",
        version="0.1.0-draft",
    )
    product_v1.ensure_input_port(
        DataProductInputPort(
            name="sales.orders",
            version=contract.version,
            contract_id=contract.id,
        )
    )
    data_product_backend.put(product_v1)

    product_v2 = product_v1.clone()
    product_v2.version = "0.2.0-draft"
    data_product_backend.put(product_v2)

    context = GovernanceReadContext(
        input_binding=DataProductInputBinding(
            data_product="Sales.KPIs",
            port_name="sales.orders",
            data_product_version="0.1.0-draft",
        ),
        allowed_data_product_statuses=("draft",),
        enforce_data_product_status=False,
    )

    plan = backend.resolve_read_context(context=context)
    assert plan.input_binding is not None
    assert plan.input_binding.data_product_version == "0.1.0-draft"

    assessment = backend.evaluate_read_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )

    backend.register_read_activity(plan=plan, assessment=assessment)


def test_resolve_write_context_prefers_contract_reference(governance_fixture):
    backend, data_product_backend, contract = governance_fixture

    context = GovernanceWriteContext(
        contract=ContractReference(
            contract_id=contract.id,
            contract_version=contract.version,
        ),
        output_binding=DataProductOutputBinding(
            data_product="dp.analytics",
            port_name="derived",
        ),
        dataset_id="analytics.orders.out",
    )

    plan = backend.resolve_write_context(context=context)
    assert plan.contract_version == contract.version
    assessment = backend.evaluate_write_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )
    with pytest.raises(RuntimeError) as excinfo:
        backend.register_write_activity(plan=plan, assessment=assessment)

    assert data_product_backend.last_output_call is not None
    assert data_product_backend.last_output_call["data_product_id"] == "dp.analytics"
    assert data_product_backend.last_output_call["port"].name == "derived"
    assert "requires review" in str(excinfo.value)


def test_register_write_activity_respects_binding_version(governance_fixture):
    backend, data_product_backend, contract = governance_fixture

    product_v1 = OpenDataProductStandard(
        id="Sales.KPIs",
        status="draft",
        version="0.1.0-draft",
    )
    product_v1.ensure_output_port(
        DataProductOutputPort(
            name="kpis.simple",
            version=contract.version,
            contract_id=contract.id,
        )
    )
    data_product_backend.put(product_v1)

    product_v2 = product_v1.clone()
    product_v2.version = "0.2.0-draft"
    data_product_backend.put(product_v2)

    context = GovernanceWriteContext(
        output_binding=DataProductOutputBinding(
            data_product="Sales.KPIs",
            port_name="kpis.simple",
            data_product_version="0.1.0-draft",
        ),
        dataset_id="sales.kpis",
        dataset_version="2024-01-01",
        allowed_data_product_statuses=("draft",),
        enforce_data_product_status=False,
    )

    plan = backend.resolve_write_context(context=context)
    assert plan.output_binding is not None
    assert plan.output_binding.data_product_version == "0.1.0-draft"

    assessment = backend.evaluate_write_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )

    backend.register_write_activity(plan=plan, assessment=assessment)


def test_get_dataset_records_returns_runs(governance_fixture):
    backend, _, contract = governance_fixture

    backend.evaluate_dataset(
        contract_id=contract.id,
        contract_version=contract.version,
        dataset_id="analytics.orders",
        dataset_version="2024-01-01",
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )

    records = backend.get_dataset_records(dataset_id="analytics.orders")

    assert records
    assert records[0]["dataset_name"] == "analytics.orders"
    assert records[0]["status"] == "ok"


def test_resolve_write_context_from_existing_output(governance_fixture):
    backend, _, contract = governance_fixture

    context = GovernanceWriteContext(
        output_binding=DataProductOutputBinding(
            data_product="dp.analytics",
            port_name="primary",
        )
    )

    plan = backend.resolve_write_context(context=context)
    assert plan.contract_id == contract.id
    assert plan.output_binding is not None

    assessment = backend.evaluate_write_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )
    backend.register_write_activity(plan=plan, assessment=assessment)


def test_register_write_skips_registration_for_pinned_version(governance_fixture):
    backend, data_product_backend, _ = governance_fixture

    product = data_product_backend.latest("dp.analytics")
    assert product is not None
    version = product.version
    assert version
    port = product.find_output_port("primary")
    assert port is not None
    port.custom_properties = [
        {"property": "dc43.output.physical_location", "value": "table"}
    ]
    data_product_backend.put(product)

    data_product_backend.last_output_call = None

    context = GovernanceWriteContext(
        output_binding=DataProductOutputBinding(
            data_product="dp.analytics",
            port_name="primary",
            data_product_version=version,
        ),
        dataset_id="analytics.orders.out",
    )

    plan = backend.resolve_write_context(context=context)
    assessment = backend.evaluate_write_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )

    backend.register_write_activity(plan=plan, assessment=assessment)

    assert data_product_backend.last_output_call is None


def test_register_read_skips_registration_for_pinned_version(governance_fixture):
    backend, data_product_backend, _ = governance_fixture

    product = data_product_backend.latest("dp.analytics")
    assert product is not None
    version = product.version
    assert version
    port = product.find_input_port("orders")
    assert port is not None
    port.custom_properties = [
        {"property": "dc43.input.refresh_mode", "value": "batch"}
    ]
    data_product_backend.put(product)

    data_product_backend.last_input_call = None

    context = GovernanceReadContext(
        input_binding=DataProductInputBinding(
            data_product="dp.analytics",
            port_name="orders",
            data_product_version=version,
        )
    )

    plan = backend.resolve_read_context(context=context)
    assessment = backend.evaluate_read_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )

    product.status = "draft"
    data_product_backend.put(product)

    with pytest.raises(ValueError, match="status"):
        backend.register_read_activity(plan=plan, assessment=assessment)

    assert data_product_backend.last_input_call is None


def test_resolve_read_context_rejects_draft_product(governance_fixture):
    backend, data_product_backend, _ = governance_fixture

    product = data_product_backend.latest("dp.analytics")
    assert product is not None
    product.status = "draft"
    data_product_backend.put(product)

    context = GovernanceReadContext(
        input_binding=DataProductInputBinding(
            data_product="dp.analytics",
            port_name="orders",
        )
    )

    with pytest.raises(ValueError, match="status"):
        backend.resolve_read_context(context=context)


def test_resolve_read_context_allows_draft_when_permitted(governance_fixture):
    backend, data_product_backend, _ = governance_fixture

    product = data_product_backend.latest("dp.analytics")
    assert product is not None
    product.status = "draft"
    data_product_backend.put(product)

    context = GovernanceReadContext(
        input_binding=DataProductInputBinding(
            data_product="dp.analytics",
            port_name="orders",
        ),
        allowed_data_product_statuses=("active", "draft"),
    )

    plan = backend.resolve_read_context(context=context)
    assert plan.policy.allowed_data_product_statuses == ("active", "draft")

    assessment = backend.evaluate_read_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )
    backend.register_read_activity(plan=plan, assessment=assessment)


def test_publish_lineage_event_registers_links(governance_fixture):
    backend, data_product_backend, contract = governance_fixture

    dataset_id = "analytics.orders"
    dataset_version = "2024-01-01"
    run_id = "44695653-fc1a-4ec6-8c2a-6c6a44ec5ad9"
    lineage_payload = {
        "eventType": "COMPLETE",
        "eventTime": "2024-01-01T00:00:00Z",
        "producer": "https://dc43.example/tests",
        "schemaURL": "https://openlineage.io/spec/2-0-2/OpenLineage.json#",
        "run": {
            "runId": run_id,
            "facets": {
                "dc43PipelineContext": {"context": {"job": "orders.read"}},
                "dc43Validation": {"status": "ok", "ok": True},
            },
        },
        "job": {"namespace": "dc43", "name": "orders-job"},
        "inputs": [
            {
                "namespace": "dc43",
                "name": dataset_id,
                "facets": {
                    "dc43Dataset": {
                        "datasetId": dataset_id,
                        "datasetVersion": dataset_version,
                        "operation": "read",
                    },
                    "dc43Contract": {
                        "contractId": contract.id,
                        "contractVersion": contract.version,
                    },
                    "dc43DataProduct": {
                        "dataProduct": "dp.analytics",
                        "portName": "orders",
                    },
                    "version": {"datasetVersion": dataset_version},
                },
            }
        ],
        "outputs": [],
    }

    event = decode_lineage_event(lineage_payload)
    assert event is not None

    backend.publish_lineage_event(event=event)

    assert data_product_backend.last_input_call is not None
    assert backend.get_linked_contract_version(
        dataset_id=dataset_id,
        dataset_version=dataset_version,
    ) == f"{contract.id}:{contract.version}"

    activity = backend.get_pipeline_activity(
        dataset_id=dataset_id,
        dataset_version=dataset_version,
    )
    assert activity, "pipeline activity should include lineage event"
    lineage_entry = activity[0].get("lineage_event")
    assert isinstance(lineage_entry, Mapping)
    assert lineage_entry.get("run", {}).get("runId") == run_id
def test_resolve_read_context_enforces_source_contract_version(governance_fixture):
    backend, _, _ = governance_fixture

    context = GovernanceReadContext(
        input_binding=DataProductInputBinding(
            data_product="dp.analytics",
            port_name="orders",
            source_data_product="dp.analytics",
            source_output_port="primary",
            source_contract_version="==9.9.9",
        )
    )

    with pytest.raises(ValueError, match="does not satisfy"):
        backend.resolve_read_context(context=context)


def test_resolve_write_context_rejects_draft_product(governance_fixture):
    backend, data_product_backend, _ = governance_fixture

    product = data_product_backend.latest("dp.analytics")
    assert product is not None
    product.status = "draft"
    data_product_backend.put(product)

    context = GovernanceWriteContext(
        output_binding=DataProductOutputBinding(
            data_product="dp.analytics",
            port_name="primary",
        )
    )

    with pytest.raises(ValueError, match="status"):
        backend.resolve_write_context(context=context)


def test_resolve_write_context_allows_draft_when_permitted(governance_fixture):
    backend, data_product_backend, _ = governance_fixture

    product = data_product_backend.latest("dp.analytics")
    assert product is not None
    product.status = "draft"
    data_product_backend.put(product)

    context = GovernanceWriteContext(
        output_binding=DataProductOutputBinding(
            data_product="dp.analytics",
            port_name="primary",
        ),
        allowed_data_product_statuses=("active", "draft"),
    )

    plan = backend.resolve_write_context(context=context)
    assert plan.policy.allowed_data_product_statuses == ("active", "draft")

    assessment = backend.evaluate_write_plan(
        plan=plan,
        validation=ValidationResult(ok=True, status="ok"),
        observations=lambda: ObservationPayload(metrics={}, schema=None),
    )
    backend.register_write_activity(plan=plan, assessment=assessment)


def test_resolve_write_context_enforces_product_version(governance_fixture):
    backend, _, _ = governance_fixture

    context = GovernanceWriteContext(
        output_binding=DataProductOutputBinding(
            data_product="dp.analytics",
            port_name="primary",
            data_product_version="==9.9.9",
        )
    )

    with pytest.raises(ValueError, match="could not be retrieved"):
        backend.resolve_write_context(context=context)


def test_pipeline_activity_uses_batch_statuses() -> None:
    class DummyStore:
        def __init__(self) -> None:
            self.matrix_calls = 0
            self.status_calls = 0

        def load_pipeline_activity(self, *, dataset_id: str, dataset_version: str | None = None):
            return (
                {
                    "dataset_id": dataset_id,
                    "dataset_version": dataset_version or "2024-01-01",
                    "contract_id": "sales.orders",
                    "contract_version": "1.0.0",
                },
            )

        def load_status_matrix_entries(
            self,
            *,
            dataset_id: str,
            dataset_versions: Sequence[str] | None = None,
            contract_ids: Sequence[str] | None = None,
        ):
            self.matrix_calls += 1
            version = (dataset_versions or ["2024-01-01"])[0]
            return (
                {
                    "dataset_id": dataset_id,
                    "dataset_version": version,
                    "contract_id": "sales.orders",
                    "contract_version": "1.0.0",
                    "status": ValidationResult(status="ok"),
                },
            )

        def load_status(self, **_kwargs):
            self.status_calls += 1
            return ValidationResult(status="warn")

    store = DummyStore()
    backend = LocalGovernanceServiceBackend(
        contract_client=object(),
        dq_client=object(),
        data_product_client=None,
        store=store,
    )

    records = backend.get_pipeline_activity(
        dataset_id="sales.orders",
        dataset_version="2024-01-01",
        include_status=True,
    )

    assert records
    assert isinstance(records[0].get("validation_status"), ValidationResult)
    assert store.matrix_calls == 1
    assert store.status_calls == 0



"""HTTP delegate implementing the data-quality backend interface."""

from __future__ import annotations

from typing import Mapping, Sequence

from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

from dc43_service_clients.data_quality import ObservationPayload, ValidationResult
from dc43_service_clients.data_quality.client.interface import DataQualityServiceClient

from .interface import DataQualityServiceBackend


class RemoteDataQualityServiceBackend(DataQualityServiceBackend):
    """Forward data-quality evaluations to a remote HTTP endpoint."""

    def __init__(self, client: DataQualityServiceClient) -> None:
        self._client = client

    def evaluate(
        self,
        *,
        contract: OpenDataContractStandard,
        payload: ObservationPayload,
    ) -> ValidationResult:
        return self._client.evaluate(contract=contract, payload=payload)

    def describe_expectations(
        self,
        *,
        contract: OpenDataContractStandard,
    ) -> Sequence[Mapping[str, object]]:
        return self._client.describe_expectations(contract=contract)

    def close(self) -> None:
        """Close the underlying client when it exposes a ``close`` hook."""

        try:
            closer = self._client.close
        except AttributeError:
            return
        closer()


__all__ = ["RemoteDataQualityServiceBackend"]




"""HTTP-based governance store proxying to remote observability services.""" 

from __future__ import annotations

from typing import Mapping, Optional, Sequence

from dc43_service_clients.data_quality import ValidationResult, coerce_details

from ._metrics import extract_metrics
from .interface import GovernanceStore

try:  # pragma: no cover - optional dependency guard
    import httpx
except ModuleNotFoundError as exc:  # pragma: no cover - surfaced via builder
    raise ModuleNotFoundError(
        "httpx is required when using the HttpGovernanceStore. Install 'dc43-service-backends[http]' or provide httpx manually.",
    ) from exc


class HttpGovernanceStore(GovernanceStore):
    """Delegate governance persistence to a remote HTTP API."""

    def __init__(
        self,
        base_url: str,
        *,
        headers: Mapping[str, str] | None = None,
        token: str | None = None,
        token_header: str = "Authorization",
        token_scheme: str = "Bearer",
        timeout: float = 10.0,
    ) -> None:
        self._base_url = base_url.rstrip("/")
        auth_headers = dict(headers or {})
        if token:
            auth_headers[token_header] = f"{token_scheme} {token}".strip()
        self._client = httpx.Client(base_url=self._base_url, headers=auth_headers, timeout=timeout)

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------
    def _status_url(self, dataset_id: str, dataset_version: str) -> str:
        return f"/datasets/{dataset_id}/versions/{dataset_version}/status"

    def _activity_url(self, dataset_id: str, dataset_version: Optional[str] = None) -> str:
        if dataset_version is None:
            return f"/datasets/{dataset_id}/activity"
        return f"/datasets/{dataset_id}/versions/{dataset_version}/activity"

    def _link_url(self, dataset_id: str, dataset_version: str) -> str:
        return f"/datasets/{dataset_id}/versions/{dataset_version}/link"

    def _metrics_url(self, dataset_id: str) -> str:
        return f"/datasets/{dataset_id}/metrics"

    def _datasets_url(self) -> str:
        return "/datasets"

    # ------------------------------------------------------------------
    # Status persistence
    # ------------------------------------------------------------------
    def save_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        status: ValidationResult | None,
    ) -> None:
        payload = None
        if status is not None:
            payload = {
                "contract_id": contract_id,
                "contract_version": contract_version,
                "status": status.status,
                "reason": status.reason,
                "details": status.details,
            }
            metrics_map = extract_metrics(status)
            if metrics_map:
                payload["metrics"] = metrics_map
        response = self._client.request(
            "DELETE" if status is None else "PUT",
            self._status_url(dataset_id, dataset_version),
            json=payload,
        )
        response.raise_for_status()

    def load_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
    ) -> ValidationResult | None:
        response = self._client.get(self._status_url(dataset_id, dataset_version))
        if response.status_code == 404:
            return None
        response.raise_for_status()
        data = response.json()
        linked_id = data.get("contract_id")
        linked_version = data.get("contract_version")
        if linked_id and linked_version and (linked_id, linked_version) != (contract_id, contract_version):
            reason = f"dataset linked to contract {linked_id}:{linked_version}"
            return ValidationResult(status="block", reason=reason, details=data)
        return ValidationResult(
            status=str(data.get("status", "unknown")),
            reason=data.get("reason"),
            details=coerce_details(data.get("details")),
        )

    # ------------------------------------------------------------------
    # Dataset links
    # ------------------------------------------------------------------
    def link_dataset_contract(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: str,
        contract_version: str,
    ) -> None:
        response = self._client.put(
            self._link_url(dataset_id, dataset_version),
            json={
                "contract_id": contract_id,
                "contract_version": contract_version,
            },
        )
        response.raise_for_status()

    def get_linked_contract_version(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> str | None:
        response = self._client.get(
            self._link_url(dataset_id, dataset_version or "latest")
            if dataset_version is not None
            else f"/datasets/{dataset_id}/link"
        )
        if response.status_code == 404:
            return None
        response.raise_for_status()
        data = response.json()
        cid = data.get("contract_id")
        cver = data.get("contract_version")
        if cid and cver:
            return f"{cid}:{cver}"
        return None

    def load_metrics(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
        contract_id: Optional[str] = None,
        contract_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        params: dict[str, str] = {}
        if dataset_version is not None:
            params["dataset_version"] = dataset_version
        if contract_id is not None:
            params["contract_id"] = contract_id
        if contract_version is not None:
            params["contract_version"] = contract_version
        response = self._client.get(self._metrics_url(dataset_id), params=params or None)
        if response.status_code == 404:
            return []
        response.raise_for_status()
        data = response.json()
        if isinstance(data, list):
            return [dict(item) for item in data if isinstance(item, Mapping)]
        if isinstance(data, Mapping):
            return [dict(data)]
        return []

    def list_datasets(self) -> Sequence[str]:
        response = self._client.get(self._datasets_url())
        if response.status_code == 404:
            return []
        response.raise_for_status()
        payload = response.json()
        if isinstance(payload, list):
            return [str(item) for item in payload]
        if isinstance(payload, Mapping):
            items = payload.get("datasets")
            if isinstance(items, list):
                return [str(item) for item in items]
        return []

    # ------------------------------------------------------------------
    # Pipeline activity
    # ------------------------------------------------------------------
    def record_pipeline_event(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        event: Mapping[str, object],
        lineage_event: Mapping[str, object] | None = None,
    ) -> None:
        payload = dict(event)
        payload.setdefault("contract_id", contract_id)
        payload.setdefault("contract_version", contract_version)
        if lineage_event is not None:
            payload["lineage_event"] = dict(lineage_event)
        response = self._client.post(self._activity_url(dataset_id, dataset_version), json=payload)
        response.raise_for_status()

    def load_pipeline_activity(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        url = (
            self._activity_url(dataset_id, dataset_version)
            if dataset_version is not None
            else self._activity_url(dataset_id)
        )
        params = {"version": dataset_version} if dataset_version is None else None
        response = self._client.get(url, params=params)
        if response.status_code == 404:
            return []
        response.raise_for_status()
        data = response.json()
        if isinstance(data, list):
            return [item for item in data if isinstance(item, Mapping)]
        if isinstance(data, Mapping):
            return [data]
        return []


__all__ = ["HttpGovernanceStore"]
